""" 
From: https://github.com/dariush-bahrami/character-tokenizer/blob/master/charactertokenizer/core.py

CharacterTokenzier for Hugging Face Transformers.
This is heavily inspired from CanineTokenizer in transformers package.
"""
import json
import os
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

from transformers.tokenization_utils import AddedToken, PreTrainedTokenizer


class CharacterTokenizer(PreTrainedTokenizer):
    def __init__(self, characters: Sequence[str], model_max_length: int, padding_side: str='left', **kwargs):
        """Character tokenizer for Hugging Face transformers.
        Args:
            characters (Sequence[str]): List of desired characters. Any character which
                is not included in this list will be replaced by a special token called
                [UNK] with id=6. Following are list of all of the special tokens with
                their corresponding ids:
                    "[CLS]": 0
                    "[SEP]": 1
                    "[BOS]": 2
                    "[MASK]": 3
                    "[PAD]": 4
                    "[RESERVED]": 5
                    "[UNK]": 6
                an id (starting at 7) will be assigned to each character.
            model_max_length (int): Model maximum sequence length.
        """
        self.characters = characters
        self.model_max_length = model_max_length
        bos_token = AddedToken("[BOS]", lstrip=False, rstrip=False)
        eos_token = AddedToken("[SEP]", lstrip=False, rstrip=False)
        sep_token = AddedToken("[SEP]", lstrip=False, rstrip=False)
        cls_token = AddedToken("[CLS]", lstrip=False, rstrip=False)
        pad_token = AddedToken("[PAD]", lstrip=False, rstrip=False)
        unk_token = AddedToken("[UNK]", lstrip=False, rstrip=False)

        mask_token = AddedToken("[MASK]", lstrip=True, rstrip=False)

        super().__init__(
            bos_token=bos_token,
            eos_token=sep_token,
            sep_token=sep_token,
            cls_token=cls_token,
            pad_token=pad_token,
            mask_token=mask_token,
            unk_token=unk_token,
            add_prefix_space=False,
            model_max_length=model_max_length,
            padding_side=padding_side,
            **kwargs,
        )

        self._vocab_str_to_int = {
            "[CLS]": 0,
            "[SEP]": 1,
            "[BOS]": 2,
            "[MASK]": 3,
            "[PAD]": 4,
            "[RESERVED]": 5,
            "[UNK]": 6,
            **{ch: i + 7 for i, ch in enumerate(characters)},
        }
        self._vocab_int_to_str = {v: k for k, v in self._vocab_str_to_int.items()}

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_str_to_int)

    def _tokenize(self, text: str) -> List[str]:
        return list(text)

    def _convert_token_to_id(self, token: str) -> int:
        return self._vocab_str_to_int.get(token, self._vocab_str_to_int["[UNK]"])

    def _convert_id_to_token(self, index: int) -> str:
        return self._vocab_int_to_str[index]

    def convert_tokens_to_string(self, tokens):
        return "".join(tokens)

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        # cls = [self.cls_token_id]
        result = token_ids_0 + sep
        if token_ids_1 is not None:
            result += token_ids_1 + sep
        return result

    def get_special_tokens_mask(
        self,
        token_ids_0: List[int],
        token_ids_1: Optional[List[int]] = None,
        already_has_special_tokens: bool = False,
    ) -> List[int]:
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=True,
            )

        result = ([0] * len(token_ids_0)) + [1]
        if token_ids_1 is not None:
            result += ([0] * len(token_ids_1)) + [1]
        return result

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        result = len(cls + token_ids_0 + sep) * [0]
        if token_ids_1 is not None:
            result += len(token_ids_1 + sep) * [1]
        return result

    def get_config(self) -> Dict:
        return {
            "char_ords": [ord(ch) for ch in self.characters],
            "model_max_length": self.model_max_length,
        }

    @classmethod
    def from_config(cls, config: Dict) -> "CharacterTokenizer":
        cfg = {}
        cfg["characters"] = [chr(i) for i in config["char_ords"]]
        cfg["model_max_length"] = config["model_max_length"]
        return cls(**cfg)

    def save_pretrained(self, save_directory: Union[str, os.PathLike], **kwargs):
        cfg_file = Path(save_directory) / "tokenizer_config.json"
        cfg = self.get_config()
        with open(cfg_file, "w") as f:
            json.dump(cfg, f, indent=4)

    @classmethod
    def from_pretrained(cls, save_directory: Union[str, os.PathLike], **kwargs):
        cfg_file = Path(save_directory) / "tokenizer_config.json"
        with open(cfg_file) as f:
            cfg = json.load(f)
        return cls.from_config(cfg)


"""Implementations of several types of Discrete Sin/Cosine Transforms with various reductions to FFT.

Currently not used by S4
"""

import torch
import torch.nn as nn
import numpy as np
import scipy.fft
from einops import rearrange, repeat

class DCT(nn.Module):
    """ Reductions adapted from https://dsp.stackexchange.com/questions/2807/fast-cosine-transform-via-fft """

    def __init__(self, N, norm='backward'):
        super().__init__()

        self.N = N

        # Materialize DCT matrix
        P = scipy.fft.dct(np.eye(N), norm=norm, type=2).T
        P = torch.tensor(P, dtype=torch.float)
        self.register_buffer('P', P)

        # TODO take care of normalization
        Q = np.exp(-1j * np.pi / (2 * self.N) * np.arange(self.N))
        Q = torch.tensor(Q, dtype=torch.cfloat)
        self.register_buffer('Q', Q) # half shift

    def forward(self, x, mode=2):
        if mode == 0:
            return self.forward_dense(x)
        elif mode == 1:
            return self.forward_n(x)
        elif mode == 2:
            return self.forward_2n(x)
        elif mode == 4:
            return self.forward_4n(x)

    def forward_dense(self, x):
        """ Baseline DCT type II - matmul by DCT matrix """
        y = (self.P.to(x) @ x.unsqueeze(-1)).squeeze(-1)
        return y

    def forward_4n(self, x):
        """ DCT type II - reduction to FFT size 4N """
        assert self.N == x.shape[-1]
        x = torch.cat([x, x.flip(-1)], dim=-1)
        z = torch.zeros_like(x)
        x = torch.stack([z, x], dim=-1)
        x = x.view(x.shape[:-2] + (-1,))
        y = torch.fft.fft(x)
        y = y[..., :self.N]
        if torch.is_complex(x):
            return y
        else:
            return torch.real(y)

    def forward_2n(self, x):
        """ DCT type II - reduction to FFT size 2N mirrored

        The reduction from the DSP forum is not quite correct in the complex input case.
        halfshift(FFT[a, b, c, d, d, c, b, a]) -> [A, B, C, D, 0, -D, -C, -B]
        In the case of real input, the intermediate step after FFT has form [A, B, C, D, 0, D*, C*, B*]
        """
        assert self.N == x.shape[-1]
        x = torch.cat([x, x.flip(-1)], dim=-1)
        y = torch.fft.fft(x)[..., :self.N]
        y = y * self.Q
        if torch.is_complex(x):
            return y
        else:
            return torch.real(y)

    def forward_n(self, x):
        """ DCT type II - reduction to size N """
        assert self.N == x.shape[-1]
        x = torch.cat([x[..., 0::2], x[..., 1::2].flip(-1)], dim=-1)
        y = torch.fft.fft(x)
        y = y * 2 * self.Q
        if torch.is_complex(x):
            y = torch.cat([y[..., :1], (y[..., 1:] + 1j * y[..., 1:].flip(-1)) / 2], dim=-1) # TODO in-place sum
        else:
            y = torch.real(y)
        return y

class IDCT(nn.Module):
    def __init__(self, N, norm='backward'):
        super().__init__()

        self.N = N

        # Materialize DCT matrix
        P = np.linalg.inv(scipy.fft.dct(np.eye(N), norm=norm, type=2).T)
        P = torch.tensor(P, dtype=torch.float)
        self.register_buffer('P', P)

        # TODO take care of normalization
        Q = np.exp(-1j * np.pi / (2 * self.N) * np.arange(2*self.N))
        Q = torch.tensor(Q, dtype=torch.cfloat)
        self.register_buffer('Q', Q) # half shift

    def forward(self, x, mode=2):
        if mode == 0:
            return self.forward_dense(x)
        elif mode == 1:
            return self.forward_n(x)
        elif mode == 2:
            return self.forward_2n(x)
        elif mode == 4:
            return self.forward_4n(x)

    def forward_dense(self, x):
        """ Baseline DCT type II - matmul by DCT matrix """
        y = (self.P.to(x) @ x.unsqueeze(-1)).squeeze(-1)
        return y

    def forward_4n(self, x):
        """ DCT type II - reduction to FFT size 4N """
        assert self.N == x.shape[-1]
        z = x.new_zeros(x.shape[:-1] + (1,))
        x = torch.cat([x, z, -x.flip(-1), -x[..., 1:], z, x[..., 1:].flip(-1)], dim=-1)
        y = torch.fft.ifft(x)
        y = y[..., 1:2*self.N:2]
        if torch.is_complex(x):
            return y
        else:
            return torch.real(y)

    def forward_2n(self, x):
        """ DCT type II - reduction to FFT size 2N mirrored """
        assert self.N == x.shape[-1]
        z = x.new_zeros(x.shape[:-1] + (1,))
        x = torch.cat([x, z, -x[..., 1:].flip(-1)], dim=-1)
        x = x / self.Q
        y = torch.fft.ifft(x)[..., :self.N]
        if torch.is_complex(x):
            return y
        else:
            return torch.real(y)

    def forward_n(self, x):
        """ DCT type II - reduction to size N """
        assert self.N == x.shape[-1]
        raise NotImplementedError # Straightforward by inverting operations of DCT-II reduction

def test_dct_ii():
    N = 8
    dct = DCT(N)

    baseline = dct.forward_dense
    methods = [dct.forward_4n, dct.forward_2n, dct.forward_n]

    # Real case
    print("DCT-II Real input")
    x = torch.randn(1, N)
    y = baseline(x)
    print(y)
    for fn in methods:
        y_ = fn(x)
        print("err", torch.norm(y-y_))

    # Complex case
    print("DCT-II Complex input")
    x = torch.randn(N) + 1j * torch.randn(N)
    y = baseline(x)
    print(y)
    for fn in methods:
        y_ = fn(x)
        print("err", torch.norm(y-y_))

def test_dct_iii():
    N = 8
    dct = IDCT(N)

    baseline = dct.forward_dense
    methods = [dct.forward_4n, dct.forward_2n]

    # Real case
    print("DCT-III Real input")
    x = torch.randn(1, N)
    y = baseline(x)
    print(y)
    for fn in methods:
        y_ = fn(x)
        print("err", torch.norm(y-y_))

    # Complex case
    print("DCT-III Complex input")
    # x = torch.randn(N) + 1j * torch.randn(N)
    x = 1j * torch.ones(N)
    y = baseline(x)
    print(y)
    for fn in methods:
        y_ = fn(x)
        print("err", torch.norm(y-y_))



#!/usr/bin/env python3
import argparse
import yaml 
from tqdm import tqdm
import typing as tp
import numpy as np
import pandas as pd
from copy import deepcopy
from collections import OrderedDict
import torch 
torch.backends.cuda.matmul.allow_tf32 = True
torch.backends.cudnn.allow_tf32 = True
import torch.nn.functional as F
import pytorch_lightning as pl
from einops import rearrange

import sys, os
FILEDIR = os.path.realpath(__file__)
sys.path.append(os.path.join(FILEDIR, '..'))
from src.models.sequence.long_conv_lm import ConvLMHeadModel
# from src.dataloaders.icl_genomics_dataloader import ICLGenomics
from src.dataloaders.genomics import ICLGenomics



# TODO: 
# Make use of maximum long context: either put entire downstream dataset in context
# or add many tunable soft tokens (soft prompting)!
# -> just fill the context up one way or another and show whats possible!



DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def instruction_tuned_ICL():

    parser = argparse.ArgumentParser()
    parser.add_argument("--ckpt_path", help="Path to pretrained model checkpoint")
    parser.add_argument("--config", default='./configs/evals/instruction_tuned_genomics.yaml')
    parser.add_argument("--results", default='./results/instruction_tuned_genomics')
    args = parser.parse_args()
    os.makedirs(args.results, exist_ok=True)

    # load configs
    config = yaml.load(open(args.config, 'r'), Loader=yaml.FullLoader)
    cfg_model = config['model'].copy()
    cfg_dataset = config['dataset'].copy()
    cfg_tuning = config['tuning'].copy()
    
    np.random.seed(config['seed'])
    torch.manual_seed(config['seed'])
    rng = np.random.RandomState(config['seed'])

    # dataset_name                                num_seqs        num_classes     median_len    std
    # dummy_mouse_enhancers_ensembl       1210            2               2381          984.4  
    # demo_coding_vs_intergenomic_seqs    100_000         2               200           0
    # demo_human_or_worm                  100_000         2               200           0
    # human_enhancers_cohn                27791           2               500           0
    # human_enhancers_ensembl             154842          2               269           122.6
    # human_ensembl_regulatory            289061          3               401           184.3
    # human_nontata_promoters             36131           2               251           0
    # human_ocr_ensembl                   174756          2               315           108.1
    
    nuc_chars = list('ACGTN')
    characters = nuc_chars # + chrom_names
    label_to_token = {0: 'A', 1: 'N'}
    datasets = {
        'human_enhancers_cohn': {
            'max_length': 502,
            'd_output': 2,
            'characters': characters,
            'label_to_token': label_to_token,
        },
        'human_nontata_promoters': {
            'max_length': 251, #253
            'd_output': 2,
            'characters': characters,
            'label_to_token': label_to_token,
        },
        'human_enhancers_ensembl': {
            'max_length': 320,
            'd_output': 2,
            'characters': characters,
            'label_to_token': label_to_token,
        },
        'human_ensembl_regulatory': {
            'max_length': 600,
            'd_output': 3,
            'characters': characters,
            'label_to_token': {0: 'A', 1: 'G', 2: 'N'},
        },
        'human_ocr_ensembl': {
            'max_length': 420,
            'd_output': 2,
            'characters': characters,
            'label_to_token': label_to_token,
        }
    }

    print('\n\nEvaluating instruction-tuned ICL performance... ')
    df_results = []
    df_i = 0
    for tuning_samples in cfg_tuning['tuning_samples']:
        print(f'...when tuning on {tuning_samples} samples...')

        for shots in cfg_dataset['shots']:
            print(f'...with {shots} shots...')
        
            for dataset, dataset_cfg in datasets.items():
                print(f'...from dataset {dataset}...')

                print(f'Collecting tuning data...')
                cfg = cfg_dataset.copy()
                cfg.update(dataset_cfg)
                cfg['dataset_name'] = dataset
                cfg['shots'] = shots
                loader = ICLGenomics(**cfg)
                loader.setup()
                
                # collect tuning samples
                tuning_X = []
                train_loader = iter(loader.train_dataloader())
                samples_collected = 0
                for x, y  in tqdm(train_loader):
                    n = min(tuning_samples, x.shape[0])
                    tuning_X.append(torch.cat([x[:n], y[:n]], dim=1))
                    samples_collected += n
                    if samples_collected >= tuning_samples:
                        print(f'...stop becuase {tuning_samples} samples collected.')
                        break
                tuning_X = torch.cat(tuning_X, dim=0)
                if shots>0:
                    tuning_y_idx = np.concatenate([
                        [cfg['max_length']+1],
                        np.arange((2*cfg['max_length'])+4, tuning_X.shape[1], cfg['max_length']+3)
                    ])
                else:
                    tuning_y_idx = cfg['max_length']+1
                tuning_y = tuning_X[:,tuning_y_idx]
                tuning_loss_mask = tuning_y_idx-1 # prediction is always from previous token

                print('Tuning pretrained model...')
                pretrained_model = load_model(cfg_model, args.ckpt_path)
                pretrained_model.to(DEVICE)
                tuned_pretrained_model = tune_model(
                    deepcopy(pretrained_model).to(DEVICE),
                    tuning_X,
                    tuning_y,
                    cfg_tuning,
                    loss_mask=tuning_loss_mask,
                    rng=rng
                )
                
                # print('Tuning untrained model...')
                # scratch_model = load_model(cfg_model)
                # scratch_model.to(DEVICE)
                # tuned_scratch_model = tune_model(
                #     scratch_model,
                #     tuning_X,
                #     tuning_y,
                #     cfg_tuning,
                #     loss_mask=tuning_loss_mask,
                #     rng=rng
                # )

                print('Evaluating ICL performance...')
                for label, model in zip(
                    ['tuned_pretrained'], #, 'scratchtrained' 
                    [tuned_pretrained_model] # tuned_scratch_model
                ):
                    print(f'{label}:')
                    acc = eval_on_loaders(model, {dataset: loader})[dataset]
                    df_results.append(
                        pd.DataFrame({
                            'dataset': dataset,
                            'tuning_samples': tuning_samples,
                            'model': label,
                            'shots': shots,
                            'eval_acc': acc
                        }, index=[df_i])
                    )
                    df_i += 1
                    pd.concat(df_results).to_csv(
                        os.path.join(args.results, 'instruction_tuned_genomics.csv')
                    )

def load_model(cfg_model, ckpt_path: str=None):

    model = ConvLMHeadModel(**cfg_model)

    if ckpt_path is not None:
        state_dict = torch.load(ckpt_path, map_location='cpu')
        # loads model from ddp by removing prexix to single if necessary
        torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(
            state_dict["state_dict"], "model."
        )
        model_state_dict = state_dict["state_dict"]
        # need to remove torchmetrics. to remove keys, need to convert to list first
        for key in list(model_state_dict.keys()):
            if "torchmetrics" in key:
                model_state_dict.pop(key)

        model.load_state_dict(model_state_dict)   
  
    return LitModel(model)


class LitModel(pl.LightningModule):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, x: torch.Tensor):
        return self.model(x)[0]
    
    def step(self, batch: tp.Tuple[torch.Tensor], loss_mask: tp.Union[int, np.ndarray]=-1, phase: str='train'):
        x, y = batch['x'].to(DEVICE), batch['y'].to(DEVICE)
        loss_mask = -1 if loss_mask is None else loss_mask
        out = self(x)
        logits = out.logits[:,loss_mask]
        if logits.ndim>2:
            logits = rearrange(logits, 'b n c -> (b n) c')
        if y.ndim==2:
            y = rearrange(y, 'b n -> (b n)')
        loss = F.cross_entropy(logits, y)
        preds = logits.argmax(axis=-1)
        acc = torch.mean((preds==y).to(torch.float32))
        return {'loss': loss, 'acc': acc}


def tune_model(model, X, y, cfg_tuning, max_epochs: int=1, loss_mask=None, verbose: bool=True, rng: np.random.RandomState=None):

    rng = np.random.RandomState(0) if rng is None else rng

    # # we use expected moving average of model for downstream ICL...
    # ema = deepcopy(model).to(DEVICE)
    # requires_grad(ema, False)
    # update_ema(ema, model, decay=0)  # Ensure EMA is initialized with synced weights
    # ema.eval()
        
    optimizer = torch.optim.AdamW(
        model.parameters(),
        weight_decay=float(cfg_tuning['weight_decay']),
        lr=float(cfg_tuning['lr'])
    )
    
    # split train/eval
    n_samples = X.shape[0]
    train_idx = np.arange(n_samples)

    batch_size = min(len(train_idx), cfg_tuning['batch_size'])
    epoch = 0
    step = 0
    losses, accs = [], []
    stop_training = False
    while not stop_training:
        if verbose:
            print(f'Epoch {epoch}...')
    
        # train epoch:
        model.train()
        rng.shuffle(train_idx)
        batch_i, batch_start = 0, 0
        while batch_start+batch_size <= len(train_idx):
            
            idx = train_idx[batch_start:batch_start+batch_size]
            batch = {'x': X[idx], 'y': y[idx]}
            model.on_train_batch_start(batch=batch, batch_idx=step)
            out = model.step(batch, loss_mask=loss_mask)
            loss, acc = out['loss'], out['acc']
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), cfg_tuning.get('gradient_clip_val', 1.0))
            losses.append(loss.cpu().detach().numpy().mean())
            accs.append(acc.cpu().detach().numpy())
            
            # accumulate gradients of N batches
            if (batch_i + 1) % cfg_tuning['accumulate_grad_batches'] == 0:
                optimizer.step()
                optimizer.zero_grad()
                # update_ema(ema, model, decay=cfg_tuning['ema_decay'])
                step += 1
                print(f'step: {step}; train loss: {losses[-1]}, acc: {accs[-1]}')
            batch_start += batch_size
            batch_i += 1
        
        epoch += 1
        if epoch>=max_epochs:
            stop_training = True

    return model #, ema


@torch.no_grad()
def update_ema(ema_model, model, decay=0.999):
    ema_params = OrderedDict(ema_model.named_parameters())
    model_params = OrderedDict(model.named_parameters())
    for name, param in model_params.items():
        ema_params[name].mul_(decay).add_(param.data, alpha=1 - decay)

def requires_grad(model, flag=True):
    for p in model.parameters():
        p.requires_grad = flag


def eval_on_loaders(model, loaders):

    results = {}
    for name, loader in loaders.items():
        print(f'Evaluating on {name} data...')

        all_acc = []
        val_loader = loader.val_dataloader()
        for batch in tqdm(val_loader):
            x, y = batch
            x = x.to(DEVICE)
            
            with torch.no_grad():
                out = model(x)
                if type(out) == tuple: out = out[0]
                logits = out.logits[:, -1]
                logits = logits.cpu().detach().numpy()
                batch_preds = logits.argmax(axis=-1)

            # batch_preds = np.array(batch_preds)
            y = y.cpu().detach().numpy()
            batch_preds = batch_preds.flatten()
            y = y.flatten()
            acc = (batch_preds == y).mean()
            all_acc.append(acc)
        
        results[name] = np.mean(all_acc)
        print(f"{name}; full eval. accuracy: {results[name]:1.4f}")

    return results


if __name__ == "__main__":
    instruction_tuned_ICL()


import numpy as np
from pytorch_lightning.callbacks import Callback

import src.utils as utils
from src.utils import registry


class ProgressiveResizing(Callback):

    def __init__(self, stage_params: list):
        """
        stage_params is a list of dicts
        e.g. stage_params = [
            {'resolution': 4, 'epochs': 50}, # 32 x 32
            {'resolution': 2, 'epochs': 30}, # 64 x 64
            {'resolution': 1, 'epochs': 20}, # 128 x 128
        ]
        """
        super().__init__()
        assert len(stage_params) > 0, 'No stages specified'
        assert all([{'resolution', 'epochs'} <= set(stage.keys()) for stage in stage_params]), \
            'stage_params must contain keys: resolution and epochs'

        self.stage_params = stage_params
        self.stage_epochs_cume = np.cumsum([stage['epochs'] for stage in stage_params])

        self._current_stage = 0

    def _verify_stages(self, trainer, model):
        # Double-check that stage parameters are correct, otherwise we'll fail in the middle of training
        for stage in self.stage_params:
            if hasattr(stage, 'scheduler'):
                # Verify that we can actually create the scheduler when we need to update it in each stage
                scheduler = utils.instantiate(registry.scheduler, {**model.hparams.scheduler, **stage['scheduler']}, trainer.optimizers[0])
                del scheduler

    def on_train_start(self, trainer, model) -> None:
        # Verify all the stage parameters are correct
        self._verify_stages(trainer, model)

        print(f"Training starts at {trainer.current_epoch}")
        if trainer.current_epoch == 0:
            # Update the model to the first stage
            self._update_to_current_stage(trainer, model)
        else:
            # Preemption or resumption of progressive resizing
            # Update the stage to the current one
            self._current_stage = int(np.searchsorted(self.stage_epochs_cume - 1, trainer.current_epoch))
            self._starting_stage = np.any(trainer.current_epoch == self.stage_epochs_cume)

            print("Progressive Resizing: Restarting at Stage {}".format(self._current_stage))
            if self._starting_stage:
                self._update_lr_scheduler(trainer, model)

            # Set the dataloader and model
            self._update_dataloaders(trainer, model)
            self._update_model(trainer, model)

        return super().on_train_start(trainer, model)

    def _update_lr_scheduler(self, trainer, model):
        if not hasattr(self.stage_params[self._current_stage], 'scheduler'):
            # No scheduler specified, so don't update the current scheduler
            return

        assert len(trainer.lr_schedulers) == 1
        # Reinitialize the scheduler
        # We don't need to carry over information from the last scheduler e.g. the last_epoch property,
        # because that will mess with the new scheduler when we step it
        hparams = {**model.hparams.scheduler, **self.stage_params[self._current_stage]['scheduler']}

        # Note that passing in the optimizer below is okay: the scheduler will be reinitialized and doesn't seem to inherit any current lr info from the optimizer
        trainer.lr_schedulers[0]['scheduler'] = utils.instantiate(registry.scheduler, hparams, trainer.optimizers[0])

        print("\tChanged scheduler to {}".format(hparams))

    def _update_dataloaders(self, trainer, model):
        # Set the train resolution and reset the dataloader
        model.hparams.loader.train_resolution = self.stage_params[self._current_stage]['resolution']
        trainer.reset_train_dataloader(model)

        print('\tChanged resolution to {}'.format(self.stage_params[self._current_stage]['resolution']))

    def _update_model(self, trainer, model):
        if not hasattr(self.stage_params[self._current_stage], 'bandlimit'):
            return

        # Update the bandlimit value for the model: this is a hack to make sure the model is updated
        # Iterate over all the modules
        for module in model.modules():
            if hasattr(module, 'bandlimit'):
                module.bandlimit = self.stage_params[self._current_stage]['bandlimit']

        print('\tChanged bandlimit to {}'.format(self.stage_params[self._current_stage]['bandlimit']))

    def _update_to_current_stage(self, trainer, model):
        print("Progressive Resizing: Moving to Stage {}".format(self._current_stage))
        # Update the train dataloader, model and scheduler
        self._update_dataloaders(trainer, model)
        self._update_model(trainer, model)
        self._update_lr_scheduler(trainer, model)


    def on_train_epoch_end(self, trainer, model):
        """
        Check to see if new stage is reached for the next epoch, and if so, prepare the new stage by
        changing the dataloader.

        (We do next epoch so that the dataloader is prepared before the next epoch)
        """
        next_epoch = trainer.current_epoch + 1

        # Check if stage should be increased
        if next_epoch >= self.stage_epochs_cume[self._current_stage] and self._current_stage < len(self.stage_params) - 1:
            self._current_stage += 1
            self._update_to_current_stage(trainer, model)

        return super().on_train_epoch_end(trainer, model)



"""Utilities for special optimizer hyperparameters.

group_parameters_for_optimizer is a modification of timm's optimizer logic, which is currently unused
add_optimizer_hooks is an improved version that uses this codebase's _optim dictionary
"""

import inspect

import torch.nn as nn

import hydra


def add_optimizer_hooks(
    model,
    bias_weight_decay=False,
    normalization_weight_decay=False,
):
    """Set weight_decay=0.0 for parameters in model.no_weight_decay, for parameters with
    attribute _no_weight_decay==True, for bias parameters if bias_weight_decay==False, for
    normalization parameters if normalization_weight_decay==False
    """

    # Separate out all parameters to those that will and won't experience regularizing weight decay
    blacklist_weight_modules = (nn.Embedding, )
    if not normalization_weight_decay:
        blacklist_weight_modules += (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
                                     # Not compatible with Pytorch 1.8.1
                                     # nn.LazyBatchNorm1d, nn.LazyBatchNorm2d, nn.LazyBatchNorm3d,
                                     nn.GroupNorm, nn.SyncBatchNorm,
                                     nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d,
                                     nn.LayerNorm, nn.LocalResponseNorm)
    for mn, m in model.named_modules():
        for pn, p in m.named_parameters():
            if (not bias_weight_decay and pn.endswith('bias')) \
                    or getattr(p, '_no_weight_decay', False) \
                    or isinstance(m, blacklist_weight_modules):
                        setattr(p, "_optim", {"weight_decay": 0.0})


def group_parameters_for_optimizer(
    model,
    optimizer_cfg,
    bias_weight_decay=False,
    normalization_weight_decay=False,
):
    """Set weight_decay=0.0 for parameters in model.no_weight_decay, for parameters with
    attribute _no_weight_decay==True, for bias parameters if bias_weight_decay==False, for
    normalization parameters if normalization_weight_decay==False
    """
    # Get the weight decay from the config, or from the default value of the optimizer constructor
    # if it's not specified in the config.
    if 'weight_decay' in optimizer_cfg:
        weight_decay = optimizer_cfg.weight_decay
    else:
        # https://stackoverflow.com/questions/12627118/get-a-function-arguments-default-value
        signature = inspect.signature(hydra.utils.get_class(optimizer_cfg._target_))
        if 'weight_decay' in signature.parameters:
            weight_decay = signature.parameters['weight_decay'].default
            if weight_decay is inspect.Parameter.empty:
                weight_decay = 0.0
        else:
            weight_decay = 0.0

    # If none of the parameters have weight decay anyway, and there are no parameters with special
    # optimization params
    if weight_decay == 0.0 and not any(hasattr(p, '_optim') for p in model.parameters()):
        return model.parameters()

    skip = model.no_weight_decay() if hasattr(model, 'no_weight_decay') else set()
    skip_keywords = (model.no_weight_decay_keywords() if hasattr(model, 'no_weight_decay_keywords')
                     else set())

    # Adapted from https://github.com/karpathy/minGPT/blob/master/mingpt/model.py#L134
    """
    This long function is unfortunately doing something very simple and is being very defensive:
    We are separating out all parameters of the model into two buckets: those that will experience
    weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
    We are then returning the PyTorch optimizer object.
    """

    # separate out all parameters to those that will and won't experience regularizing weight decay
    decay = set()
    no_decay = set()
    special = set()
    whitelist_weight_modules = (nn.Linear, )
    blacklist_weight_modules = (nn.Embedding, )
    if not normalization_weight_decay:
        blacklist_weight_modules += (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d,
                                     # Not compatible with Pytorch 1.8.1
                                     # nn.LazyBatchNorm1d, nn.LazyBatchNorm2d, nn.LazyBatchNorm3d,
                                     nn.GroupNorm, nn.SyncBatchNorm,
                                     nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d,
                                     nn.LayerNorm, nn.LocalResponseNorm)
    for mn, m in model.named_modules():
        for pn, p in m.named_parameters():
            fpn = '%s.%s' % (mn, pn) if mn else pn # full param name
            if not p.requires_grad:
                continue  # frozen weights
            if hasattr(p, '_optim'):
                special.add(fpn)
            elif fpn in skip or any(skip_keyword in fpn for skip_keyword in skip_keywords):
                no_decay.add(fpn)
            elif getattr(p, '_no_weight_decay', False):
                no_decay.add(fpn)
            elif not bias_weight_decay and pn.endswith('bias'):
                no_decay.add(fpn)
            elif pn.endswith('weight') and isinstance(m, whitelist_weight_modules):
                # weights of whitelist modules will be weight decayed
                decay.add(fpn)
            elif isinstance(m, blacklist_weight_modules):
                # weights of blacklist modules will NOT be weight decayed
                no_decay.add(fpn)

    param_dict = {pn: p for pn, p in model.named_parameters() if p.requires_grad}
    # special case the position embedding parameter in the root GPT module as not decayed
    if 'pos_emb' in param_dict:
        no_decay.add('pos_emb')

    # In case of parameter sharing, some parameters show up in decay but are not in param_dict.keys()
    decay &= param_dict.keys()
    decay |= (param_dict.keys() - no_decay - special)
    # validate that we considered every parameter
    inter_params = decay & no_decay
    union_params = decay | no_decay
    assert len(inter_params) == 0, f"Parameters {str(inter_params)} made it into both decay/no_decay sets!"
    assert len(param_dict.keys() - special - union_params) == 0, f"parameters {str(param_dict.keys() - union_params)}  were not separated into either decay/no_decay set!"

    if weight_decay == 0.0 or not no_decay:
        param_groups = [{"params": [param_dict[pn] for pn in sorted(list(no_decay | decay))],
                         "weight_decay": weight_decay}]
    else:
        param_groups = [
            {"params": [param_dict[pn] for pn in sorted(list(decay))], "weight_decay": weight_decay},
            {"params": [param_dict[pn] for pn in sorted(list(no_decay))], "weight_decay": 0.0},
        ]
    # Add parameters with special hyperparameters
    # Unique dicts
    hps = [dict(s) for s in set(frozenset(param_dict[pn]._optim.items()) for pn in special)]
    for hp in hps:
        params = [param_dict[pn] for pn in sorted(list(special)) if param_dict[pn]._optim == hp]
        param_groups.append({"params": params, **hp})

    return param_groups



from itertools import islice
from functools import partial
import os
import functools
# import json
# from pathlib import Path
# from pyfaidx import Fasta
# import polars as pl
# import pandas as pd
import torch
from random import randrange, random
import numpy as np
from pathlib import Path


from src.dataloaders.datasets.hg38_char_tokenizer import CharacterTokenizer
from genomic_benchmarks.loc2seq import download_dataset
from genomic_benchmarks.data_check import is_downloaded
from src.dataloaders.base import default_data_path

"""

Genomic Benchmarks Dataset, from:
https://github.com/ML-Bioinfo-CEITEC/genomic_benchmarks


"""


# helper functions

def exists(val):
    return val is not None

def identity(t):
    return t

def cast_list(t):
    return t if isinstance(t, list) else [t]

def coin_flip():
    return random() > 0.5

# genomic function transforms

seq_indices_embed = torch.zeros(256).long()
seq_indices_embed[ord('a')] = 0
seq_indices_embed[ord('c')] = 1
seq_indices_embed[ord('g')] = 2
seq_indices_embed[ord('t')] = 3
seq_indices_embed[ord('n')] = 4
seq_indices_embed[ord('A')] = 0
seq_indices_embed[ord('C')] = 1
seq_indices_embed[ord('G')] = 2
seq_indices_embed[ord('T')] = 3
seq_indices_embed[ord('N')] = 4
seq_indices_embed[ord('.')] = -1

one_hot_embed = torch.zeros(256, 4)
one_hot_embed[ord('a')] = torch.Tensor([1., 0., 0., 0.])
one_hot_embed[ord('c')] = torch.Tensor([0., 1., 0., 0.])
one_hot_embed[ord('g')] = torch.Tensor([0., 0., 1., 0.])
one_hot_embed[ord('t')] = torch.Tensor([0., 0., 0., 1.])
one_hot_embed[ord('n')] = torch.Tensor([0., 0., 0., 0.])
one_hot_embed[ord('A')] = torch.Tensor([1., 0., 0., 0.])
one_hot_embed[ord('C')] = torch.Tensor([0., 1., 0., 0.])
one_hot_embed[ord('G')] = torch.Tensor([0., 0., 1., 0.])
one_hot_embed[ord('T')] = torch.Tensor([0., 0., 0., 1.])
one_hot_embed[ord('N')] = torch.Tensor([0., 0., 0., 0.])
one_hot_embed[ord('.')] = torch.Tensor([0.25, 0.25, 0.25, 0.25])

reverse_complement_map = torch.Tensor([3, 2, 1, 0, 4]).long()

def torch_fromstring(seq_strs):
    batched = not isinstance(seq_strs, str)
    seq_strs = cast_list(seq_strs)
    np_seq_chrs = list(map(lambda t: np.fromstring(t, dtype = np.uint8), seq_strs))
    seq_chrs = list(map(torch.from_numpy, np_seq_chrs))
    return torch.stack(seq_chrs) if batched else seq_chrs[0]

def str_to_seq_indices(seq_strs):
    seq_chrs = torch_fromstring(seq_strs)
    return seq_indices_embed[seq_chrs.long()]

def str_to_one_hot(seq_strs):
    seq_chrs = torch_fromstring(seq_strs)
    return one_hot_embed[seq_chrs.long()]

def seq_indices_to_one_hot(t, padding = -1):
    is_padding = t == padding
    t = t.clamp(min = 0)
    one_hot = F.one_hot(t, num_classes = 5)
    out = one_hot[..., :4].float()
    out = out.masked_fill(is_padding[..., None], 0.25)
    return out

# augmentations

string_complement_map = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}

def string_reverse_complement(seq):
    rev_comp = ''
    for base in seq[::-1]:
        if base in string_complement_map:
            rev_comp += string_complement_map[base]
        # if bp not complement map, use the same bp
        else:
            rev_comp += base
    return rev_comp

def seq_indices_reverse_complement(seq_indices):
    complement = reverse_complement_map[seq_indices.long()]
    return torch.flip(complement, dims = (-1,))

def one_hot_reverse_complement(one_hot):
    *_, n, d = one_hot.shape
    assert d == 4, 'must be one hot encoding with last dimension equal to 4'
    return torch.flip(one_hot, (-1, -2))



class GenomicBenchmarkDataset(torch.utils.data.Dataset):

    '''
    Loop thru bed file, retrieve (chr, start, end), query fasta file for sequence.
    Returns a generator that retrieves the sequence.
    '''

    def __init__(
        self,
        split,
        max_length,
        dataset_name="human_nontata_promoters",
        d_output=2, # default binary classification
        dest_path=None,
        tokenizer=None,
        tokenizer_name=None,
        use_padding=None,
        add_eos=False,
        rc_aug=False,
        return_augs=False,
        return_mask=False,
    ):

        self.max_length = max_length
        self.use_padding = use_padding
        self.tokenizer_name = tokenizer_name
        self.tokenizer = tokenizer
        self.return_augs = return_augs
        self.add_eos = add_eos
        self.d_output = d_output  # needed for decoder to grab
        self.rc_aug = rc_aug
        self.return_mask = return_mask

        if not is_downloaded(dataset_name, cache_path=dest_path):
            print("downloading {} to {}".format(dataset_name, dest_path))
            download_dataset(dataset_name, version=0, dest_path=dest_path)
        else:
            print("already downloaded {}-{}".format(split, dataset_name))

        # change "val" split to "test".  No val available, just test
        if split == "val":
            split = "test"

        # use Path object
        base_path = Path(dest_path) / dataset_name / split

        self.all_seqs = []
        self.all_labels = []
        label_mapper = {}

        for i, x in enumerate(base_path.iterdir()):
            label_mapper[x.stem] = i

        for label_type in label_mapper.keys():
            for path in (base_path / label_type).iterdir():
                with open(path, "r") as f:
                    content = f.read()
                self.all_seqs.append(content)
                self.all_labels.append(label_mapper[label_type])
                
    def __len__(self):
        return len(self.all_labels)

    def __getitem__(self, idx):
        x = self.all_seqs[idx]
        y = self.all_labels[idx]

        # apply rc_aug here if using
        if self.rc_aug and coin_flip():
            x = string_reverse_complement(x)

        seq = self.tokenizer(x,
            add_special_tokens=True if self.add_eos else False,  # this is what controls adding eos
            padding="max_length" if self.use_padding else "do_not_pad",
            max_length=self.max_length,
            truncation=True,
        )
        seq_ids = seq["input_ids"]  # get input_ids

        seq_ids = torch.LongTensor(seq_ids)

        # need to wrap in list
        target = torch.LongTensor([y])  # offset by 1, includes eos

        if self.return_mask:
            return seq_ids, target, {'mask': torch.BoolTensor(seq['attention_mask'])}
        else:
            return seq_ids, target






if __name__ == '__main__':
    """Quick test loading dataset.
    
    example
    python -m src.dataloaders.datasets.genomic_bench_dataset
    
    """

    max_length = 300  # max len of seq grabbed
    use_padding = True
    dest_path = "data/genomic_benchmark/"
    return_mask = True
    add_eos = True
    padding_side = 'right'    

    tokenizer = CharacterTokenizer(
        characters=['A', 'C', 'G', 'T', 'N'],
        model_max_length=max_length,
        add_special_tokens=False,
        padding_side=padding_side,
    )

    ds = GenomicBenchmarkDataset(
        max_length = max_length,
        use_padding = use_padding,
        split = 'train', # 
        tokenizer=tokenizer,
        tokenizer_name='char',
        dest_path=dest_path,
        return_mask=return_mask,
        add_eos=add_eos,
    )

    # it = iter(ds)
    # elem = next(it)
    # print('elem[0].shape', elem[0].shape)
    # print(elem)
    # breakpoint()


"""Implements downsampling and upsampling on sequences."""

import torch
from torch import nn
import torch.nn.functional as F
from einops import rearrange, repeat, reduce

from src.models.sequence import SequenceModule
from src.models.nn import LinearActivation

""" Simple pooling functions that just downsample or repeat

stride: Subsample on the layer dimension
expand: Repeat on the feature dimension
"""


class DownSample(SequenceModule):
    def __init__(self, d_input, stride=1, expand=1, transposed=True):
        super().__init__()
        self.d_input = d_input
        self.stride = stride
        self.expand = expand
        self.transposed = transposed

    def forward(self, x):
        if x is None: return None
        if self.stride > 1:
            assert x.ndim == 3, "Downsampling with higher-dimensional inputs is currently not supported. It is recommended to use average or spectral pooling instead."
            if self.transposed:
                x = x[..., 0::self.stride]
            else:
                x = x[..., 0::self.stride, :]

        if self.expand > 1:
            if self.transposed:
                x = repeat(x, 'b d ... -> b (d e) ...', e=self.expand)
            else:
                x = repeat(x, 'b ... d -> b ... (d e)', e=self.expand)

        return x, None


    def step(self, x, state, **kwargs):
        if self.stride > 1 or self.expand > 1:
            raise NotImplementedError
        return x, state

    @property
    def d_output(self):
        return self.d_input * self.expand

class DownAvgPool(SequenceModule):
    def __init__(self, d_input, stride=1, expand=None, transposed=True):
        super().__init__()
        self.d_input = d_input
        self.stride = stride
        self.expand = expand
        self.transposed = transposed

        if self.expand is not None:
            self.linear = LinearActivation(
                d_input,
                d_input * expand,
                transposed=transposed,
            )

    def forward(self, x):
        if not self.transposed:
            x = rearrange(x, 'b ... d -> b d ...')

        if self.stride > 1:
            # einops appears slower than F
            if x.ndim == 3:
                x = F.avg_pool1d(x, self.stride, self.stride)
            elif x.ndim == 4:
                x = F.avg_pool2d(x, self.stride, self.stride)
            else:
                # Reduction string e.g. "b d (l1 2) (l2 2) -> b d l1 l2"
                reduce_str = "b d " + " ".join([f"(l{i} {self.stride})" for i in range(x.ndim-2)]) \
                        + " -> b d " + " ".join([f"l{i}" for i in range(x.ndim-2)])
                x = reduce(x, reduce_str, 'mean')

        # if self.expand > 1:
        #     x = repeat(x, 'b d ... -> b (d e) ...', e=self.expand)

        if not self.transposed:
            x = rearrange(x, 'b d ... -> b ... d')
        if self.expand is not None:
            x = self.linear(x)
        return x, None

    def step(self, x, state, **kwargs):
        if self.stride > 1 or self.expand > 1:
            raise NotImplementedError
        return x, state

    @property
    def d_output(self):
        if self.expand is None:
            return self.d_input
        else:
            return self.d_input * self.expand

class DownSpectralPool(SequenceModule):
    def __init__(self, d_input, stride=1, expand=1, transposed=True):
        super().__init__()
        self.d_input = d_input
        self.stride = stride
        self.expand = expand
        self.transposed = transposed

    def forward(self, x):
        """
        x: (B, L..., D)
        """
        if not self.transposed:
            x = rearrange(x, 'b ... d -> b d ...')
        shape = x.shape[2:]
        x_f = torch.fft.ifftn(x, s=shape)

        for axis, l in enumerate(shape):
            assert l % self.stride == 0, 'input length must be divisible by stride'
            new_l = l // self.stride
            idx = torch.cat([torch.arange(0, new_l-new_l//2), l+torch.arange(-new_l//2, 0)]).to(x_f.device)
            x_f = torch.index_select(x_f, 2+axis, idx)
        x = torch.fft.ifftn(x_f, s=[l//self.stride for l in shape])
        x = x.real

        if self.expand > 1:
            x = repeat(x, 'b d ... -> b (d e) ...', e=self.expand)
        if not self.transposed:
            x = rearrange(x, 'b d ... -> b ... d')
        return x, None

    def step(self, x, state, **kwargs):
        if self.stride > 1 or self.expand > 1:
            raise NotImplementedError
        return x, state

    @property
    def d_output(self):
        return self.d_input * self.expand

class UpSample(SequenceModule):
    def __init__(self, d_input, stride=1, expand=1, transposed=True):
        super().__init__()
        self.d_input = d_input
        self.stride = stride
        self.expand = expand
        self.transposed = transposed

    def forward(self, x):
        if x is None: return None
        if self.expand > 1:
            if self.transposed:
                x = reduce(x, '... (d e) l -> ... d l', 'mean', e=self.expand)
            else:
                x = reduce(x, '... (d e) -> ... d', 'mean', e=self.expand)
        if self.stride > 1:
            if self.transposed:
                x = repeat(x, '... l -> ... (l e)', e=self.stride)
            else:
                x = repeat(x, '... l d -> ... (l e) d', e=self.stride)
        return x, None

    @property
    def d_output(self):
        return self.d_input // self.expand
    def step(self, x, state, **kwargs):
        if self.stride > 1 or self.expand > 1:
            raise NotImplementedError
        return x, state
class UpAvgPool(SequenceModule):
    def __init__(self, d_input, stride=1, expand=1, causal=False, transposed=True):
        super().__init__()
        assert d_input % expand == 0
        self.d_input = d_input
        self.stride = stride
        self.expand = expand
        self.causal = causal
        self.transposed = transposed

        self.linear = LinearActivation(
            d_input,
            d_input // expand,
            transposed=transposed,
        )

    def forward(self, x):
        # TODO only works for 1D right now
        if x is None: return None
        x = self.linear(x)
        if self.stride > 1:
            if self.transposed:
                if self.causal:
                    x = F.pad(x[..., :-1], (1, 0)) # Shift to ensure causality
                x = repeat(x, '... l -> ... (l e)', e=self.stride)
            else:
                if self.causal:
                    x = F.pad(x[..., :-1, :], (0, 0, 1, 0)) # Shift to ensure causality
                x = repeat(x, '... l d -> ... (l e) d', e=self.stride)
        return x, None

    @property
    def d_output(self):
        return self.d_input // self.expand
    def step(self, x, state, **kwargs):
        if self.stride > 1 or self.expand > 1:
            raise NotImplementedError
        return x, state

class DownLinearPool(SequenceModule):
    def __init__(self, d_model, stride=1, expand=1, causal=False, transposed=True):
        super().__init__()

        self.d_model = d_model
        self.stride = stride
        self.expand = expand
        self.transposed = transposed

        self.linear = LinearActivation(
            d_model * stride,
            d_model * expand,
            transposed=transposed,
        )

    def forward(self, x):
        if self.transposed:
            x = rearrange(x, '... h (l s) -> ... (h s) l', s=self.stride)
        else:
            x = rearrange(x, '... (l s) h -> ... l (h s)', s=self.stride)
        x = self.linear(x)
        return x, None

    def step(self, x, state, **kwargs):
        # if self.stride > 1 or self.expand > 1:
        #     raise NotImplementedError
        # return x, state
        if x is None: return None, state
        state.append(x)
        if len(state) == self.stride:
            x = rearrange(torch.stack(state, dim=-1), '... h s -> ... (h s)')
            if self.transposed: x = x.unsqueeze(-1)
            x = self.linear(x)
            if self.transposed: x = x.squeeze(-1)
            return x, []
        else:
            return None, state

    def default_state(self, *batch_shape, device=None):
        return []

    @property
    def d_output(self):
        return self.d_input * self.expand

class UpLinearPool(SequenceModule):
    def __init__(self, d, stride=1, expand=1, causal=False, transposed=True):
        super().__init__()

        # self.d_model = d * expand
        # self.d_output = d
        assert d % expand == 0
        self.d_model = d
        self.d_output = d // expand
        # self._d_output = d_output
        self.stride = stride
        self.causal = causal
        self.transposed = transposed

        self.linear = LinearActivation(
            self.d_model,
            self.d_output * stride,
            transposed=transposed,
        )

    def forward(self, x, skip=None):
        x = self.linear(x)
        if self.transposed:
            if self.causal:
                x = F.pad(x[..., :-1], (1, 0)) # Shift to ensure causality
            x = rearrange(x, '... (h s) l -> ... h (l s)', s=self.stride)
        else:
            if self.causal:
                x = F.pad(x[..., :-1, :], (0, 0, 1, 0)) # Shift to ensure causality
            x = rearrange(x, '... l (h s) -> ... (l s) h', s=self.stride)
        if skip is not None:
            x = x + skip
        return x, None

    def step(self, x, state, **kwargs):
        """
        x: (..., H)
        """

        assert len(state) > 0
        y, state = state[0], state[1:]
        if len(state) == 0:
            assert x is not None
            if self.transposed: x = x.unsqueeze(-1)
            x = self.linear(x)
            if self.transposed: x = x.squeeze(-1)
            x = rearrange(x, '... (h s) -> ... h s', s=self.stride)
            state = list(torch.unbind(x, dim=-1))
        else: assert x is None
        return y, state

    def default_state(self, *batch_shape, device=None):
        state = torch.zeros(batch_shape + (self.d_output, self.stride), device=device) # (batch, h, s)
        state = list(torch.unbind(state, dim=-1)) # List of (..., H)
        return state

    # @property
    # def d_output(self): return self._d_output

""" Pooling functions with trainable parameters """ # TODO make d_output expand instead

class DownPool2d(SequenceModule):

    def __init__(self, d_input, d_output, stride=1, transposed=True, weight_norm=True):
        super().__init__()

        self.linear = LinearActivation(
            d_input,
            d_output,
            transposed=transposed,
            weight_norm=weight_norm,
        )

        self.pool = nn.AvgPool2d(kernel_size=stride, stride=stride),

    def forward(self, x):
        if self.transposed:
            x = self.pool(x)

# TODO DownPool/UpPool are currently used by unet/sashimi backbones
# DownLinearPool is used by the registry (for isotropic backbone)
# DownPool is essentially the same as DownLinearPool. These should be consolidated
class DownPool(SequenceModule):
    def __init__(self, d_input, d_output=None, expand=None, stride=1, transposed=True, weight_norm=True, initializer=None, activation=None):
        super().__init__()
        assert (d_output is None) + (expand is None) == 1
        if d_output is None: d_output = d_input * expand

        self.d_output = d_output
        self.stride = stride
        self.transposed = transposed

        self.linear = LinearActivation(
            d_input * stride,
            d_output,
            transposed=transposed,
            initializer=initializer,
            weight_norm = weight_norm,
            activation=activation,
            activate=True if activation is not None else False,
        )

    def forward(self, x):
        if self.transposed:
            x = rearrange(x, '... h (l s) -> ... (h s) l', s=self.stride)
        else:
            x = rearrange(x, '... (l s) h -> ... l (h s)', s=self.stride)
        x = self.linear(x)
        return x, None

    def step(self, x, state, **kwargs):
        """
        x: (..., H)
        """

        if x is None: return None, state
        state.append(x)
        if len(state) == self.stride:
            x = rearrange(torch.stack(state, dim=-1), '... h s -> ... (h s)')
            if self.transposed: x = x.unsqueeze(-1)
            x = self.linear(x)
            if self.transposed: x = x.squeeze(-1)
            return x, []
        else:
            return None, state

    def default_state(self, *batch_shape, device=None):
        return []


class UpPool(SequenceModule):
    def __init__(self, d_input, d_output, stride, transposed=True, weight_norm=True, initializer=None, activation=None):
        super().__init__()

        self.d_input = d_input
        self._d_output = d_output
        self.stride = stride
        self.transposed = transposed

        self.linear = LinearActivation(
            d_input,
            d_output * stride,
            transposed=transposed,
            initializer=initializer,
            weight_norm = weight_norm,
            activation=activation,
            activate=True if activation is not None else False,
        )

    def forward(self, x, skip=None):
        x = self.linear(x)
        if self.transposed:
            x = F.pad(x[..., :-1], (1, 0)) # Shift to ensure causality
            x = rearrange(x, '... (h s) l -> ... h (l s)', s=self.stride)
        else:
            x = F.pad(x[..., :-1, :], (0, 0, 1, 0)) # Shift to ensure causality
            x = rearrange(x, '... l (h s) -> ... (l s) h', s=self.stride)
        if skip is not None:
            x = x + skip
        return x, None

    def step(self, x, state, **kwargs):
        """
        x: (..., H)
        """

        assert len(state) > 0
        y, state = state[0], state[1:]
        if len(state) == 0:
            assert x is not None
            if self.transposed: x = x.unsqueeze(-1)
            x = self.linear(x)
            if self.transposed: x = x.squeeze(-1)
            x = rearrange(x, '... (h s) -> ... h s', s=self.stride)
            state = list(torch.unbind(x, dim=-1))
        else: assert x is None
        return y, state

    def default_state(self, *batch_shape, device=None):
        state = torch.zeros(batch_shape + (self.d_output, self.stride), device=device) # (batch, h, s)
        state = list(torch.unbind(state, dim=-1)) # List of (..., H)
        return state

    @property
    def d_output(self): return self._d_output

registry = {
    'sample': DownSample,
    'pool': DownAvgPool,
    'avg': DownAvgPool,
    'linear': DownLinearPool,
    'spectral': DownSpectralPool,
}

up_registry = {
    # 'sample': UpSample,
    'pool': UpAvgPool,
    'avg': UpAvgPool,
    'linear': UpLinearPool,
    # 'spectral': UpSpectralPool, # Not implemented and no way to make this causal
}




# Copyright (c) 2019-2020, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import os
from collections import Counter
from collections import OrderedDict

import torch

import src.utils as utils


class Vocab(object):
    def __init__(self, special=[], min_freq=0, max_size=None, lower_case=True,
                 delimiter=None, vocab_file=None):
        self.counter = Counter()
        self.special = special
        self.min_freq = min_freq
        self.max_size = max_size
        self.lower_case = lower_case
        self.delimiter = delimiter
        self.vocab_file = vocab_file

    def tokenize(self, line, add_eos=False, add_double_eos=False):
        line = line.strip()
        # convert to lower case
        if self.lower_case:
            line = line.lower()

        # empty delimiter '' will evaluate False
        if self.delimiter == '':
            symbols = line
        else:
            symbols = line.split(self.delimiter)

        if add_double_eos:  # lm1b
            return ['<S>'] + symbols + ['<S>']
        elif add_eos:
            return symbols + ['<eos>']
        else:
            return symbols

    def count_file(self, path, verbose=False, add_eos=False):
        if verbose:
            print('counting file {} ...'.format(path))
        assert os.path.exists(path)

        sents = []
        with open(path, 'r', encoding='utf-8') as f:
            for idx, line in enumerate(f):
                if verbose and idx > 0 and idx % 500000 == 0:
                    print('    line {}'.format(idx))
                symbols = self.tokenize(line, add_eos=add_eos)
                self.counter.update(symbols)
                sents.append(symbols)

        return sents

    def count_sents(self, sents, verbose=False):
        """
            sents : a list of sentences, each a list of tokenized symbols
        """
        if verbose:
            print('counting {} sents ...'.format(len(sents)))
        for idx, symbols in enumerate(sents):
            if verbose and idx > 0 and idx % 500000 == 0:
                print('    line {}'.format(idx))
            self.counter.update(symbols)

    def _build_from_file(self, vocab_file):
        self.idx2sym = []
        self.sym2idx = OrderedDict()

        with open(vocab_file, 'r', encoding='utf-8') as f:
            for line in f:
                symb = line.strip().split()[0]
                self.add_symbol(symb)
        self.unk_idx = self.sym2idx['<UNK>']

    def build_vocab(self):
        if self.vocab_file:
            print('building vocab from {}'.format(self.vocab_file))
            self._build_from_file(self.vocab_file)
            print('final vocab size {}'.format(len(self)))
        else:
            print('building vocab with min_freq={}, max_size={}'.format(
                self.min_freq, self.max_size))
            self.idx2sym = []
            self.sym2idx = OrderedDict()

            for sym in self.special:
                self.add_special(sym)

            for sym, cnt in self.counter.most_common(self.max_size):
                if cnt < self.min_freq:
                    break
                self.add_symbol(sym)

            print('final vocab size {} from {} unique tokens'.format(
                len(self), len(self.counter)))

    def encode_file(self, path, ordered=False, verbose=False, add_eos=True,
                    add_double_eos=False):
        if verbose:
            print('encoding file {} ...'.format(path))
        assert os.path.exists(path)
        encoded = []
        with open(path, 'r', encoding='utf-8') as f:
            for idx, line in enumerate(f):
                if verbose and idx > 0 and idx % 500000 == 0:
                    print('    line {}'.format(idx))
                symbols = self.tokenize(line, add_eos=add_eos,
                                        add_double_eos=add_double_eos)
                encoded.append(self.convert_to_tensor(symbols))

        if ordered:
            encoded = torch.cat(encoded)

        return encoded

    def encode_sents(self, sents, ordered=False, verbose=False):
        if verbose:
            print('encoding {} sents ...'.format(len(sents)))
        encoded = []
        for idx, symbols in enumerate(sents):
            if verbose and idx > 0 and idx % 500000 == 0:
                print('    line {}'.format(idx))
            encoded.append(self.convert_to_tensor(symbols))

        if ordered:
            encoded = torch.cat(encoded)

        return encoded

    def add_special(self, sym):
        if sym not in self.sym2idx:
            self.idx2sym.append(sym)
            self.sym2idx[sym] = len(self.idx2sym) - 1
            setattr(self, '{}_idx'.format(sym.strip('<>')), self.sym2idx[sym])

    def add_symbol(self, sym):
        if sym not in self.sym2idx:
            self.idx2sym.append(sym)
            self.sym2idx[sym] = len(self.idx2sym) - 1

    def get_sym(self, idx):
        assert 0 <= idx < len(self), 'Index {} out of range'.format(idx)
        return self.idx2sym[idx]

    def get_idx(self, sym):
        if sym in self.sym2idx:
            return self.sym2idx[sym]
        else:
            # print('encounter unk {}'.format(sym))
            assert '<eos>' not in sym
            assert hasattr(self, 'unk_idx')
            return self.sym2idx.get(sym, self.unk_idx)

    def get_symbols(self, indices):
        return [self.get_sym(idx) for idx in indices]

    def get_indices(self, symbols):
        return [self.get_idx(sym) for sym in symbols]

    def convert_to_tensor(self, symbols):
        return torch.LongTensor(self.get_indices(symbols))

    def convert_to_sent(self, indices, exclude=None):
        if exclude is None:
            return ' '.join([self.get_sym(idx) for idx in indices])
        else:
            return ' '.join([self.get_sym(idx) for idx in indices if idx not in exclude])

    def __len__(self):
        return len(self.idx2sym)


# Class OpenAIVocab has been adapted from
# https://github.com/cybertronai/transformer-xl/blob/master/utils/vocabulary.py
class OpenAIVocab(Vocab):
    def __init__(self, max_size=None, vocab_file=None):
        from transformers import GPT2Tokenizer
        self.tokenizer = GPT2Tokenizer.from_pretrained('gpt2')
        self.EOT = self.tokenizer.encoder['<|endoftext|>']
        self.max_size = max_size
        self.vocab_file = vocab_file

        pad = 8
        vocab_size = len(self.tokenizer)
        padded_vocab_size = (vocab_size + pad - 1) // pad * pad
        for i in range(0, padded_vocab_size - vocab_size):
            token = f'madeupword{i:09d}'
            self.tokenizer.add_tokens([token])

    def __len__(self):
        return len(self.tokenizer)

    def count_file(self, path, verbose=False, add_eos=False):
        # TODO: train from scratch, respect self.max_size
        pass

    def build_vocab(self):
        pass

    def encode_file(self, path, ordered=False, verbose=False, add_eos=True, add_double_eos=False) -> torch.LongTensor:
        cached = path + '.bpe'
        if os.path.exists(cached):
            return torch.load(cached)
        print(f'encoding file {path} ...')
        assert os.path.exists(path), f"{path} doesn't exist"

        with open(path, encoding='utf-8') as f:
            # Suppress warnings about length.
            with open(os.devnull, "w") as devnull, contextlib.redirect_stderr(devnull):
                out = torch.LongTensor(self.tokenizer.encode(f.read()) + [self.EOT])
                with utils.distributed.sync_workers() as rank:
                    if rank == 0:
                        torch.save(out, cached)
                return out

    def tokenize(self, line, add_eos=False, add_double_eos=False):
        return self.tokenizer.encode(line)

    def convert_to_tensor(self, symbols):
        return torch.LongTensor(symbols)



"""pykeops implementations of the Vandermonde matrix multiplication kernel used in the S4D kernel."""
import math
import torch

from einops import rearrange, repeat
from opt_einsum import contract

import os

try:
    import pykeops
    from pykeops.torch import LazyTensor, Genred
except:
    pass

try:
    from cauchy_mult import vand_log_mult_sym_fwd, vand_log_mult_sym_bwd
except:
    vand_log_mult_sym_fwd, vand_log_mult_sym_bwd = None, None

_conj = lambda x: torch.cat([x, x.conj()], dim=-1)
def _broadcast_dims(*tensors):
    max_dim = max([len(tensor.shape) for tensor in tensors])
    tensors = [tensor.view((1,)*(max_dim-len(tensor.shape))+tensor.shape) for tensor in tensors]
    return tensors

def _c2r(x): return torch.view_as_real(x)
def _r2c(x): return torch.view_as_complex(x)

def vandermonde_naive(v, x, L, conj=True):
    """
    v: (..., N)
    x: (..., N)
    returns: (..., L) \sum v x^l
    """
    if conj:
        x = _conj(x)
        v = _conj(v)
    vandermonde_matrix = x.unsqueeze(-1) ** torch.arange(L).to(x) # (... N L)
    vandermonde_prod = torch.sum(v.unsqueeze(-1) * vandermonde_matrix, dim=-2) # (... L)
    return vandermonde_prod

def log_vandermonde_naive(v, x, L, conj=True):
    """
    v: (..., N)
    x: (..., N)
    returns: (..., L) \sum v x^l
    """
    vandermonde_matrix = torch.exp(x.unsqueeze(-1) * torch.arange(L).to(x)) # (... N L)
    vandermonde_prod = contract('... n, ... n l -> ... l', v, vandermonde_matrix) # (... L)
    if conj:
        return 2*vandermonde_prod.real
    else:
        return vandermonde_prod

def log_vandermonde_lazy(v, x, L, conj=True):
    if conj:
        v = _conj(v)
        x = _conj(x)
    l = torch.arange(L).to(x)
    v, x, l = _broadcast_dims(v, x, l)
    v_l = LazyTensor(rearrange(v, '... N -> ... N 1 1'))
    x_l = LazyTensor(rearrange(x, '... N -> ... N 1 1'))
    l_l = LazyTensor(rearrange(l, '... L -> ... 1 L 1'))
    # exp
    vand = (x_l * l_l).exp()
    s = (v_l*vand).sum(dim=len(v_l.shape)-2)
    return s.squeeze(-1)

def log_vandermonde(v, x, L, conj=True):
    expr = 'ComplexMult(v, ComplexExp(ComplexMult(x, l)))'
    vandermonde_mult = Genred(
        expr,
        [
            'v = Vj(2)',
            'x = Vj(2)',
            'l = Vi(2)',
        ],
        reduction_op='Sum',
        axis=1,
    )

    l = torch.arange(L).to(x)
    v, x, l = _broadcast_dims(v, x, l)
    v = _c2r(v)
    x = _c2r(x)
    l = _c2r(l)

    r = vandermonde_mult(v, x, l, backend='GPU')
    if conj:
        return 2*_r2c(r).real
    else:
        return _r2c(r)

def log_vandermonde_transpose_naive(u, v, x, L):
    vandermonde_matrix = torch.exp(x.unsqueeze(-1) * torch.arange(L).to(x)) # (... N L)
    vandermonde_prod = contract('... l, ... n, ... n l -> ... n', u.to(x), v.to(x), vandermonde_matrix) # (... L)
    return vandermonde_prod

def log_vandermonde_transpose(u, v, x, L):
    """
    u: ... H L
    v: ... H N
    x: ... H N
    Returns: ... H N

    V = Vandermonde(a, L) : (H N L)
    contract_L(V * u * v)
    """
    expr = 'ComplexMult(ComplexMult(v, u), ComplexExp(ComplexMult(x, l)))'
    vandermonde_mult = Genred(
        expr,
        [
            'u = Vj(2)',
            'v = Vi(2)',
            'x = Vi(2)',
            'l = Vj(2)',
        ],
        reduction_op='Sum',
        axis=1,
    )

    l = torch.arange(L).to(x)
    u, v, x, l = _broadcast_dims(u, v, x, l)
    u = _c2r(u)
    v = _c2r(v)
    x = _c2r(x)
    l = _c2r(l)

    r = vandermonde_mult(u, v, x, l, backend='GPU')
    return _r2c(r)

def _log_vandermonde_matmul(x, L):
    vandermonde_matrix = torch.exp(x.unsqueeze(-1) * torch.arange(L).to(x)) # (... N L)
    return vandermonde_matrix

def log_vandermonde_matmul(v, K):
    prod = contract('...n, ...nl -> ...l', v, K)
    return 2*prod.real

class LogVandMultiplySymmetric(torch.autograd.Function):

    @staticmethod
    def forward(ctx, v, x, L):
        batch, N = v.shape
        supported_N_values = [1 << log_n for log_n in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]]
        if not N in supported_N_values:
            raise NotImplementedError(f'Only support N values in {supported_N_values}')
        max_L_value = 32 * 1024 * 64 * 1024
        if L > max_L_value:
            raise NotImplementedError(f'Only support L values <= {max_L_value}')
        if not v.is_cuda and x.is_cuda:
            raise NotImplementedError(f'Only support CUDA tensors')
        ctx.save_for_backward(v, x)
        return vand_log_mult_sym_fwd(v, x, L)

    @staticmethod
    def backward(ctx, dout):
        v, x = ctx.saved_tensors
        dv, dx = vand_log_mult_sym_bwd(v, x, dout)
        return dv, dx, None


if vand_log_mult_sym_fwd and vand_log_mult_sym_bwd is not None:
    log_vandermonde_fast = LogVandMultiplySymmetric.apply
else:
    log_vandermonde_fast = None


import torch
import csv
import pandas as pd
import numpy as np
from tqdm import tqdm
import liftover
from pathlib import Path
from pyfaidx import Fasta
from random import randrange, random


def exists(val):
    return val is not None

def coin_flip():
    return random() > 0.5

string_complement_map = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}

def string_reverse_complement(seq):
    rev_comp = ''
    for base in seq[::-1]:
        if base in string_complement_map:
            rev_comp += string_complement_map[base]
        # if bp not complement map, use the same bp
        else:
            rev_comp += base
    return rev_comp

class FastaInterval():
    def __init__(
        self,
        *,
        fasta_file,
        # max_length = None,
        return_seq_indices = False,
        shift_augs = None,
        rc_aug = False
    ):
        fasta_file = Path(fasta_file)
        assert fasta_file.exists(), 'path to fasta file must exist'

        self.seqs = Fasta(str(fasta_file))
        self.return_seq_indices = return_seq_indices
        # self.max_length = max_length # -1 for adding sos or eos token
        self.shift_augs = shift_augs
        self.rc_aug = rc_aug

        # calc len of each chromosome in fasta file, store in dict
        self.chr_lens = {}

        for chr_name in self.seqs.keys():
            # remove tail end, might be gibberish code
            # truncate_len = int(len(self.seqs[chr_name]) * 0.9)
            # self.chr_lens[chr_name] = truncate_len
            self.chr_lens[chr_name] = len(self.seqs[chr_name])


    def __call__(self, chr_name, start, end, max_length, return_augs = False):
        """
        max_length passed from dataset, not from init
        """
        interval_length = end - start
        chromosome = self.seqs[chr_name]
        # chromosome_length = len(chromosome)
        chromosome_length = self.chr_lens[chr_name]

        if exists(self.shift_augs):
            min_shift, max_shift = self.shift_augs
            max_shift += 1

            min_shift = max(start + min_shift, 0) - start
            max_shift = min(end + max_shift, chromosome_length) - end

            rand_shift = randrange(min_shift, max_shift)
            start += rand_shift
            end += rand_shift

        left_padding = right_padding = 0

        # checks if not enough sequence to fill up the start to end
        if interval_length < max_length:
            extra_seq = max_length - interval_length

            extra_left_seq = extra_seq // 2
            extra_right_seq = extra_seq - extra_left_seq

            start -= extra_left_seq
            end += extra_right_seq

        if start < 0:
            left_padding = -start
            start = 0

        if end > chromosome_length:
            right_padding = end - chromosome_length
            end = chromosome_length

        # Added support!  need to allow shorter seqs
        if interval_length > max_length:
            end = start + max_length

        seq = str(chromosome[start:end])

        if self.rc_aug and coin_flip():
            seq = string_reverse_complement(seq)

        seq = ('.' * left_padding) + seq + ('.' * right_padding)

        return seq


class ChromatinProfileDataset(torch.utils.data.Dataset):
    '''
    Recreation of chromatin profile prediction benchmark from BigBird paper https://arxiv.org/abs/2007.14062
    Original sequence coordinates and target labels are provided via a csv.
    Original sequences have a length of 1000. This is changed to be max_length on the fly.
    Target labels are read into a LongTensor. Coordinates are read into a DataFrame with columns "Chr_No" (0-based), "Start" and "End".
    Original coordinates are in hg19 format named as train_hg19_coords_targets.csv etc. 
    Hg19 coordinates will be translated to hg38 if ref_genome_version=='hg38'. 
    The translated coordinated can be saved to a new file e.g. train_hg19_coords_targets.csv so this only needs to be done once.
    Returns a generator that retrieves the sequence.
    '''
    def __init__(
        self,
        max_length,
        ref_genome_path=None,
        ref_genome_version=None,
        coords_target_path=None,
        tokenizer=None,
        tokenizer_name=None,
        use_padding=None,
        add_eos=False,
        return_seq_indices=False,
        shift_augs=None,
        rc_aug=False,
        return_augs=False,
        save_liftover=False,
    ):

        self.max_length      = max_length
        assert max_length%2==0 # check window is divisible by 2
        self.use_padding     = use_padding
        self.tokenizer_name  = tokenizer_name
        self.tokenizer       = tokenizer
        self.return_augs     = return_augs
        self.add_eos         = add_eos
        self.rc_aug          = rc_aug
        self.ref_genome_version = ref_genome_version
        
        # self.ref_genome = FastaInterval(fasta_file=ref_genome_path, max_length=self.max_length)
        self.ref_genome = FastaInterval(fasta_file=ref_genome_path)
        
        # Original data coordinates are from hg19. 
        # If ref genome is hg38 and original coordinates are provided these must be translated by liftover.
        # Conversion only needs to be done once so save liftover coordinates to file optionally.
        if self.ref_genome_version=='hg19':
            if 'hg19' in coords_target_path.split('/')[-1]:
                self.load_csv_data(coords_target_path)
            else:
                raise ValueError('Make sure data coordinates are in hg19 format (and put "hg19" in filename)')
        elif self.ref_genome_version=='hg38':
            if 'hg38' in coords_target_path.split('/')[-1]:
                self.load_csv_data(coords_target_path)
            elif 'hg19' in coords_target_path.split('/')[-1]:
                self.load_csv_data(coords_target_path)
                print('ref_genome_version = "hg38" but target coordinates are labelled "hg19"')
                self.convert_coordinates(coords_target_path, save_liftover)
            else:
                raise ValueError('Make sure data coordinates have correct hg19/hg38 in filename')
        else:
            raise ValueError('ref_genome_version must be "hg19" or "hg38"')
             
        # Move start/end to new window 
        # Window = 1000 used in raw coordinate data
        self.coords['Start'] = self.coords['Start']-int((max_length-1000)/2)
        self.coords['End']   = self.coords['End']+int((max_length-1000)/2)
        
    def load_csv_data(self, coords_target_path):
        # Grab sequence coordinates from csv
        self.coords = pd.read_csv(
            coords_target_path, 
            usecols=['Chr_No','Start','End'],
            dtype={'Chr_No':np.int64,'Start':np.int64,'End':np.int64}
        ).reset_index(drop=True) # Note Chr_No is zero-based
        
        # Quickly grab target column names
        with open(coords_target_path, "r") as f:
            reader = csv.reader(f)
            header = next(reader)
            self.target_columns = [col for col in header if col[:2]=='y_' ]
            
        # Grab targets from csv and convert to torch long format
        self.targets = torch.from_numpy(
                    pd.read_csv(
                        coords_target_path, 
                        usecols=self.target_columns,
                        dtype={k:bool for k in self.target_columns}
                    ).to_numpy()
                ).long()

    def __len__(self):
        return len(self.coords)

    def __getitem__(self, idx):
        y = self.targets[idx]
        coord = self.coords.iloc[idx]
        seq = self.ref_genome(
            'chr{}'.format(coord['Chr_No']+1), # Make chromosome id 1-based 
            coord['Start'], 
            coord['End'],
            max_length=self.max_length,
        )
        
        # # apply rc_aug here if using
        # if self.rc_aug and coin_flip():
        #     seq = string_reverse_complement(seq)
        
        if self.tokenizer==None:
            return seq, y
        
        x = self.tokenizer(seq.upper()) # Apply upper() incase ref genome is soft masked
        x = torch.LongTensor(x["input_ids"]) # Grab input ids and convert to LongTensorx
        return x, y
    
    
    def convert_coordinates(self, coords_target_path, save_liftover):
        '''
        Loop through coordinates and translate from hg19 to hg38.
        Filter entries where liftover fails.
        Save this to file so we only have to do it once.
        '''
        converter = liftover.get_lifter('hg19', 'hg38')
        
        print("Translating coordinates from hg19 to hg38:")
        for i in tqdm(range(len(self.coords))):
            row = self.coords.iloc[i]
            new_start = converter['chr{}'.format(row['Chr_No']+1)][row['Start']]
            new_end = converter['chr{}'.format(row['Chr_No']+1)][row['End']]
            if (len(new_start) == 0) or (len(new_end) == 0):
                # If liftover fails set -999 for filtering
                self.coords.iloc[i]['Start']=-999
            else:
                self.coords.iloc[i]['Start']=new_start[0][1]
                self.coords.iloc[i]['End']=new_end[0][1]
                
        # Filter unmapped coordinates
        n_before = len(self.coords)
        self.coords = self.coords.query('Start!=-999')
        n_after = len(self.coords)
        print('Filtered {} unmapped coordinates. There are {} samples remaining'.format(n_before-n_after, n_after))
        
        # Filter incorrect window sizes
        n_before=n_after
        self.coords = self.coords.query('End-Start==1000')
        n_after = len(self.coords)
        print('Filtered {} incorrect window sizes. There are {} samples remaining'.format(n_before-n_after, n_after))
        
        # Reindex targets based on filtered coordinates and reset coordinate index
        self.targets = self.targets[self.coords.index.to_numpy()]
        self.coords.reset_index(inplace=True, names=['filter_index'])
        
        assert len(self.targets) == len(self.coords) # Sanity check
        
        
        if save_liftover: # save liftover coords in original format and change filename accordingly
            hg38_coords_targets = pd.concat([self.coords, pd.DataFrame(columns=self.target_columns, data=self.targets)], axis=1)
            print('Saving translated and filtered data to {}'.format(coords_target_path.replace('hg19','hg38')))
            hg38_coords_targets.to_csv(coords_target_path.replace('hg19','hg38'))
            del hg38_coords_targets


import subprocess
from typing import List, Dict
import math
from functools import cmp_to_key

import os
import vitax_rag

from importlib import resources

def get_resource_path(relative_path):
    return str(resources.files("vitax_rag").joinpath(relative_path))




_WEIGHTS = {"id": 0.6, "cov": 0.5, "bits": 0.5, "evalue": 0.4, "length": 0.3}
_DEFAULT_TIE_TOL = 0.005  #
def split_string(input_string, chunk_size=2000, step_size=00):

    chunks = []
    length = len(input_string)


    if length < chunk_size:
        return [input_string]

    for i in range(0, length - chunk_size + 1, step_size):
        chunks.append(input_string[i:i + chunk_size])

    if (length - chunk_size) % step_size != 0:
        chunks.append(input_string[-chunk_size:])

    return chunks

def reverse_complement(dna):
    
    complement = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', "N": "N"}
    complemented = ''.join(complement[base] if base in complement else base for base in dna)
    return complemented[::-1]




def run_blast(
    file_path: str,
    blast_path: str,
    db_path: str,
    outfmt: str = "6 qseqid sseqid sstart send pident length mismatch gapopen bitscore evalue qcovs",
    blast_exe: str = "blastn",
) -> bool:
    result = subprocess.run(
        [blast_exe, "-query", file_path, "-db", db_path, "-out", blast_path, "-outfmt", outfmt],
        capture_output=True,
        text=True,
    )
    return result.returncode == 0




def find_blast(gene,id, blast_path,length = 4000, bitscore = None, evalue = None):
    l = len(gene)
    with open(blast_path,"r") as f:
        for line in f:
            columns = line.strip().split("\t")
            targetid = columns[0]
            if targetid != id:
                continue
            bs = columns[4]
            evalue = columns[5]
            sseqid = columns[1].split(".")[0]  # 目标序列的ID (sseqid)
            length2 = length - len(gene)
           
            file_name = "./data/sequence/"+sseqid +".fasta"
            seq = getse(file_name=file_name)
            length = len(seq)
            send = int(columns[3])
            seq = seq[send:]
            gene = gene+seq
            break
    return gene
            
            


def find_blast_augument(gene,id, blast_path,length = 4000, bitscore = None, evalue = None):
    l = len(gene)
    with open(blast_path,"r") as f:
        flag = 1
        rows = []
        for line in f:
            columns = line.strip().split("\t")
            targetid = columns[0]
            if targetid != id and flag == 1:
                continue
            elif targetid != id and flag == 2:
                break
            rows.append(line)
    sorted_lines = rank_blast_lines(rows)
    if len(sorted_lines) == 0:
        return find_blast(gene,id, blast_path,length)
    columns = sorted_lines[0].strip().split("\t")
    length2 = length - len(gene)
    sseqid = columns[0].split(".")[0] 
    send = int(columns[2])
    
    file_name = "./data/sequence/"+sseqid +".fasta"
    try:
       seq = getse(file_name=file_name)
    except Exception:
        return find_blast(gene,id, blast_path,length)
    seq = seq[send:send + length2]
    gene = gene+seq
    return gene

def getse(file_name):
  with open(file_name, "r") as file:
    data = []
    result = ""
    lines = file.readlines()
    for  line in lines:
        if line.startswith(">"):
            continue
        data.append(line.replace("\n", ""))
    result = result.join(data)
    # print(result)
    return result

WEIGHTS = {"id": 0.6, "cov": 0.5, "bits": 0.5, "evalue": 0.4, "length": 0.3}
_DEFAULT_TIE_TOL = 0.005  # 分数近似并列的相对阈值（0.5%）


def _safe_float(x: str) -> float:
    try:
        return float(x)
    except Exception:
        return 0.0


def _safe_int(x: str) -> int:
    try:
        return int(float(x))
    except Exception:
        return 0


def _parse_line(line: str) -> Dict[str, str]:
    parts = line.strip().split("\t")
    if len(parts) < 11:
        raise ValueError("BLAST line has fewer than 11 fields")
    return {
        "qseqid": parts[0],
        "sseqid": parts[1],
        "sstart": parts[2],
        "send": parts[3],
        "pident": parts[4],
        "length": parts[5],
        "mismatch": parts[6],
        "gapopen": parts[7],
        "bitscore": parts[8],
        "evalue": parts[9],
        "qcovs": parts[10],
    }


def _compute_batch_stats(rows: List[Dict[str, str]]) -> Dict[str, float]:
    max_bits = 1.0
    max_len = 1.0
    max_ev_strength = 1.0
    for r in rows:
        bits = _safe_float(r["bitscore"])
        length = _safe_int(r["length"])
        evalue = _safe_float(r["evalue"])
        ev_strength = max(0.0, min(200.0, -math.log10(evalue + 1e-200)))
        if bits > max_bits:
            max_bits = bits
        if float(length) > max_len:
            max_len = float(length)
        if ev_strength > max_ev_strength:
            max_ev_strength = ev_strength
    return {"max_bitscore": max_bits, "max_length": max_len, "max_ev_strength": max_ev_strength}


def _synergy_score(row: Dict[str, str], stats: Dict[str, float]) -> float:
    pident = _safe_float(row["pident"])
    qcovs = _safe_float(row["qcovs"])
    bits = _safe_float(row["bitscore"])
    evalue = _safe_float(row["evalue"])
    length = float(_safe_int(row["length"]))
    mismatch = float(_safe_int(row["mismatch"]))
    gapopen = float(_safe_int(row["gapopen"]))

    id_norm = max(0.0, min(1.0, pident / 100.0))
    cov_norm = max(0.0, min(1.0, qcovs / 100.0))
    bits_norm = bits / max(stats["max_bitscore"], 1e-9)
    ev_strength = max(0.0, min(200.0, -math.log10(evalue + 1e-200)))
    ev_norm = ev_strength / max(stats["max_ev_strength"], 1e-9)
    len_norm = length / max(stats["max_length"], 1e-9)

    core = (
        (id_norm ** _WEIGHTS["id"])
        * (cov_norm ** _WEIGHTS["cov"])
        * (bits_norm ** _WEIGHTS["bits"])
        * (ev_norm ** _WEIGHTS["evalue"])
        * (len_norm ** _WEIGHTS["length"])
    )

    mismatch_rate = mismatch / max(length, 1.0)
    penalty_gap = 1.0 / (1.0 + gapopen)
    penalty_mismatch = math.exp(-2.0 * mismatch_rate)
    return core * penalty_gap * penalty_mismatch


def _tie_break_compare(a: Dict[str, str], b: Dict[str, str]) -> int:
    pa = _safe_float(a["pident"]); pb = _safe_float(b["pident"])
    if pa != pb: return -1 if pa > pb else 1

    qa = _safe_float(a["qcovs"]); qb = _safe_float(b["qcovs"])
    if qa != qb: return -1 if qa > qb else 1

    ga = _safe_int(a["gapopen"]); gb = _safe_int(b["gapopen"])
    if ga != gb: return -1 if ga < gb else 1

    ma = _safe_int(a["mismatch"]); mb = _safe_int(b["mismatch"])
    if ma != mb: return -1 if ma < mb else 1

    la = _safe_int(a["length"]); lb = _safe_int(b["length"])
    if la != lb: return -1 if la > lb else 1

    sa = a["sseqid"]; sb = b["sseqid"]
    if sa < sb: return -1
    if sa > sb: return 1
    return 0


def _compare_with_tie_tol(a: Dict[str, str], b: Dict[str, str], tie_tol: float) -> int:
    sa = a["_score"]; sb = b["_score"]
    ref = max(sa, sb, 1e-9)
    if abs(sa - sb) > tie_tol * ref:
        return -1 if sa > sb else 1
    return _tie_break_compare(a, b)


def rank_blast_lines(lines: List[str], tie_tol: float = _DEFAULT_TIE_TOL) -> List[str]:
    """
    输入：每一行 BLAST 结果（TSV 11 列）的字符串列表
    输出：按协同同源评分排序后的 'sseqid\\tsstart\\tsend' 列表
    """
    rows = []
    for line in lines:
        if not line.strip():
            continue
        try:
            rows.append(_parse_line(line))
        except ValueError:
            # 跳过非法行
            continue

    if not rows:
        return []

    stats = _compute_batch_stats(rows)
    for r in rows:
        r["_score"] = _synergy_score(r, stats)

    rows_sorted = sorted(rows, key=cmp_to_key(lambda a, b: _compare_with_tie_tol(a, b, tie_tol)))

    out = [f"{r['sseqid']}\t{r['sstart']}\t{r['send']}" for r in rows_sorted]
    return out



"""Custom learning rate schedulers"""

import math
import warnings
import torch

from timm.scheduler import CosineLRScheduler


# https://pytorch.org/docs/stable/_modules/torch/optim/lr_scheduler.html
class CosineWarmup(torch.optim.lr_scheduler.CosineAnnealingLR):

    def __init__(self, optimizer, T_max, eta_min=0, warmup_step=0, **kwargs):
        self.warmup_step = warmup_step
        super().__init__(optimizer, T_max - warmup_step, eta_min, *kwargs)

    # Copied from CosineAnnealingLR, but adding warmup and changing self.last_epoch to
    # self.last_epoch - self.warmup_step.
    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning)

        if self.last_epoch == self.warmup_step:  # also covers the case where both are 0
            return self.base_lrs
        elif self.last_epoch < self.warmup_step:
            return [base_lr * (self.last_epoch + 1) / self.warmup_step for base_lr in self.base_lrs]
        elif (self.last_epoch - self.warmup_step - 1 - self.T_max) % (2 * self.T_max) == 0:
            return [group['lr'] + (base_lr - self.eta_min) *
                    (1 - math.cos(math.pi / self.T_max)) / 2
                    for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)]
        return [(1 + math.cos(math.pi * (self.last_epoch - self.warmup_step) / self.T_max)) /
                (1 + math.cos(math.pi * (self.last_epoch - self.warmup_step - 1) / self.T_max)) *
                (group['lr'] - self.eta_min) + self.eta_min
                for group in self.optimizer.param_groups]

    _get_closed_form_lr = None


def InvSqrt(optimizer, warmup_step):
    """ Originally used for Transformer (in Attention is all you need)
    """

    def lr_lambda(step):
        # return a multiplier instead of a learning rate
        if step == warmup_step:  # also covers the case where both are 0
            return 1.
        else:
            return 1. / (step ** 0.5) if step > warmup_step else (step + 1) / (warmup_step ** 1.5)

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)


def Constant(optimizer, warmup_step):

    def lr_lambda(step):
        if step == warmup_step:  # also covers the case where both are 0
            return 1.
        else:
            return 1. if step > warmup_step else (step + 1) / warmup_step

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)


class TimmCosineLRScheduler(CosineLRScheduler, torch.optim.lr_scheduler._LRScheduler):
    """ Wrap timm.scheduler.CosineLRScheduler so we can call scheduler.step() without passing in epoch.
    It supports resuming as well.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_epoch = -1
        self.step(epoch=0)

    def step(self, epoch=None):
        if epoch is None:
            self._last_epoch += 1
        else:
            self._last_epoch = epoch
        # We call either step or step_update, depending on whether we're using the scheduler every
        # epoch or every step.
        # Otherwise, lightning will always call step (i.e., meant for each epoch), and if we set
        # scheduler interval to "step", then the learning rate update will be wrong.
        if self.t_in_epochs:
            super().step(epoch=self._last_epoch)
        else:
            super().step_update(num_updates=self._last_epoch)




#@title Huggingface Pretrained Wrapper

"""

This is script is a simple HuggingFace wrapper around a HyenaDNA model, to enable a one click example
of how to load the pretrained weights and get embeddings.

It will instantiate a HyenaDNA model (model class is in the `standalone_hyenadna.py`), and handle the downloading of pretrained weights from HuggingFace.

Check out the colab notebook for a simpler and more complete walk through of how to use HyenaDNA with pretrained weights.

"""


import json
import os
import subprocess
import torch
# import transformers
from transformers import PreTrainedModel
import re
from standalone_hyenadna import HyenaDNAModel
from standalone_hyenadna import CharacterTokenizer


# helper 1
def inject_substring(orig_str):
    """Hack to handle matching keys between models trained with and without
    gradient checkpointing."""

    # modify for mixer keys
    pattern = r"\.mixer"
    injection = ".mixer.layer"

    modified_string = re.sub(pattern, injection, orig_str)

    # modify for mlp keys
    pattern = r"\.mlp"
    injection = ".mlp.layer"

    modified_string = re.sub(pattern, injection, modified_string)

    return modified_string

# helper 2
def load_weights(scratch_dict, pretrained_dict, checkpointing=False):
    """Loads pretrained (backbone only) weights into the scratch state dict."""

    # loop thru state dict of scratch
    # find the corresponding weights in the loaded model, and set it

    # need to do some state dict "surgery"
    for key, value in scratch_dict.items():
        if 'backbone' in key:
            # the state dicts differ by one prefix, '.model', so we add that
            key_loaded = 'model.' + key
            # breakpoint()
            # need to add an extra ".layer" in key
            if checkpointing:
                key_loaded = inject_substring(key_loaded)
            try:
                scratch_dict[key] = pretrained_dict[key_loaded]
            except:
                raise Exception('key mismatch in the state dicts!')

    # scratch_dict has been updated
    return scratch_dict

class HyenaDNAPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """
    base_model_prefix = "hyenadna"

    def __init__(self, config):
        pass

    def forward(self, input_ids, **kwargs):
        return self.model(input_ids, **kwargs)

    @classmethod
    def from_pretrained(cls,
                        path,
                        model_name,
                        download=False,
                        config=None,
                        device='cpu',
                        use_head=False,
                        n_classes=2,
                      ):
        # first check if it is a local path
        pretrained_model_name_or_path = os.path.join(path, model_name)
        if os.path.isdir(pretrained_model_name_or_path) and download == False:
            if config is None:
                config = json.load(open(os.path.join(pretrained_model_name_or_path, 'config.json')))
        else:
            hf_url = f'https://huggingface.co/LongSafari/{model_name}'

            subprocess.run(f'rm -rf {pretrained_model_name_or_path}', shell=True)
            command = f'mkdir -p {path} && cd {path} && git lfs install && git clone {hf_url}'
            subprocess.run(command, shell=True)

            if config is None:
                config = json.load(open(os.path.join(pretrained_model_name_or_path, 'config.json')))

        scratch_model = HyenaDNAModel(**config, use_head=use_head, n_classes=n_classes)  # the new model format
        loaded_ckpt = torch.load(
            os.path.join(pretrained_model_name_or_path, 'weights.ckpt'),
            map_location=torch.device(device)
        )

        # need to load weights slightly different if using gradient checkpointing
        if config.get("checkpoint_mixer", False):
            checkpointing = config["checkpoint_mixer"] == True or config["checkpoint_mixer"] == True
        else:
            checkpointing = False

        # grab state dict from both and load weights
        state_dict = load_weights(scratch_model.state_dict(), loaded_ckpt['state_dict'], checkpointing=checkpointing)

        # scratch model has now been updated
        scratch_model.load_state_dict(state_dict)
        print("Loaded pretrained weights ok!")
        return scratch_model




####################################################################################################




"""# Inference (450k to 1M tokens)!

If all you're interested in is getting embeddings on long DNA sequences
(inference), then we can do that right here in Colab!


*   We provide an example how to load the weights from Huggingface.
*   On the free tier, which uses a
T4 GPU w/16GB of memory, we can process 450k tokens / nucleotides.
*   For processing 1M tokens, you'll need an A100, which Colab offers as a paid tier.
*   (Don't forget to run the entire notebook above too)

--

To pretrain or fine-tune the 1M long sequence model (8 layers, d_model=256),
you'll need 8 A100s 80GB, and all that code is in the main repo!
"""

#@title Single example
import json
import os
import subprocess
# import transformers
from transformers import PreTrainedModel

def inference_single():

    '''
    this selects which backbone to use, and grabs weights/ config from HF
    4 options:
      'hyenadna-tiny-1k-seqlen'   # fine-tune on colab ok
      'hyenadna-small-32k-seqlen'
      'hyenadna-medium-160k-seqlen'  # inference only on colab
      'hyenadna-medium-450k-seqlen'  # inference only on colab
      'hyenadna-large-1m-seqlen'  # inference only on colab
    '''

    # you only need to select which model to use here, we'll do the rest!
    pretrained_model_name = 'hyenadna-small-32k-seqlen'

    max_lengths = {
        'hyenadna-tiny-1k-seqlen': 1024,
        'hyenadna-small-32k-seqlen': 32768,
        'hyenadna-medium-160k-seqlen': 160000,
        'hyenadna-medium-450k-seqlen': 450000,  # T4 up to here
        'hyenadna-large-1m-seqlen': 1_000_000,  # only A100 (paid tier)
    }

    max_length = max_lengths[pretrained_model_name]  # auto selects

    # data settings:
    use_padding = True
    rc_aug = False  # reverse complement augmentation
    add_eos = False  # add end of sentence token

    # we need these for the decoder head, if using
    use_head = False
    n_classes = 2  # not used for embeddings only

    # you can override with your own backbone config here if you want,
    # otherwise we'll load the HF one in None
    backbone_cfg = None

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    print("Using device:", device)

    # instantiate the model (pretrained here)
    if pretrained_model_name in ['hyenadna-tiny-1k-seqlen',
                                 'hyenadna-small-32k-seqlen',
                                 'hyenadna-medium-160k-seqlen',
                                 'hyenadna-medium-450k-seqlen',
                                 'hyenadna-large-1m-seqlen']:
        # use the pretrained Huggingface wrapper instead
        model = HyenaDNAPreTrainedModel.from_pretrained(
            './checkpoints',
            pretrained_model_name,
            download=True,
            config=backbone_cfg,
            device=device,
            use_head=use_head,
            n_classes=n_classes,
        )

    # from scratch
    elif pretrained_model_name is None:
        model = HyenaDNAModel(**backbone_cfg, use_head=use_head, n_classes=n_classes)

    # create tokenizer
    tokenizer = CharacterTokenizer(
        characters=['A', 'C', 'G', 'T', 'N'],  # add DNA characters, N is uncertain
        model_max_length=max_length + 2,  # to account for special tokens, like EOS
        add_special_tokens=False,  # we handle special tokens elsewhere
        padding_side='left', # since HyenaDNA is causal, we pad on the left
    )

    #### Single embedding example ####

    # create a sample 450k long, prepare
    sequence = 'ACTG' * int(max_length/4)
    tok_seq = tokenizer(sequence)
    tok_seq = tok_seq["input_ids"]  # grab ids

    # place on device, convert to tensor
    tok_seq = torch.LongTensor(tok_seq).unsqueeze(0)  # unsqueeze for batch dim
    tok_seq = tok_seq.to(device)

    # prep model and forward
    model.to(device)
    model.eval()
    with torch.inference_mode():
        embeddings = model(tok_seq)

    print(embeddings.shape)  # embeddings here!

# # uncomment to run! (to get embeddings)
inference_single()


# to run this, just call:
    # python huggingface.py



import math
import torch
import torch.nn.functional as F
from sklearn.metrics import f1_score, roc_auc_score
from functools import partial
import torchmetrics.functional as tm_f
import torch.distributions as dist
from sklearn.metrics import f1_score, roc_auc_score, matthews_corrcoef
from torchmetrics import Metric
from torchmetrics.classification import MulticlassRecall, MulticlassPrecision

class CorrectAggregatedMetric(Metric):
    """This is needed to calculate some metrics b/c small batch sizes cause aggregation via a simple
        average to be off, as some classes might not be present in batch but will get penalized with a 0."""
    def __init__(self, class_idx: int, dist_sync_on_step=False):
        # call `self.add_state`for every internal state that is needed for the metrics computations
        # dist_reduce_fx indicates the function that should be used to reduce
        # state from multiple processes
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.class_idx = torch.tensor(class_idx)
        self.add_state("numerator", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("denominator", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def _update(self, numerator, denominator, preds, y) -> tuple:
        raise NotImplemented

    def update(self, logits: torch.Tensor, y: torch.Tensor):
        # update metric states
        preds = torch.argmax(logits, dim=-1)
        logits = logits.view(-1, logits.shape[-1])
        y = y.view(-1)
        assert preds.shape == y.shape, f"preds shape {preds.shape} != y shape {y.shape}"
        self.numerator, self.denominator = self._update(self.numerator, self.denominator, preds, y)

    def compute(self):
        # compute final result
        value = self.numerator.float() / self.denominator if self.denominator > 0 else torch.tensor(0.0)
        return value

    def reset(self):
        self.numerator = torch.tensor(0.0)
        self.denominator = torch.tensor(0.0)

class AccuracyPerClass(CorrectAggregatedMetric):
    """Calculate per class accuracy, i.e. P(y_hat = class_idx AND y = class_idx OR y_hat != class_idx AND y != class_idx)
    """
    def _update(self, numerator, denominator, preds, y) -> tuple:
        # Filter down to the class of interest
        class_idx = self.class_idx
        relevant_idxs = (y == class_idx)
        numerator += (preds[relevant_idxs] == class_idx).sum()
        denominator += relevant_idxs.sum()
        relevant_idxs = (y != class_idx)
        numerator += (preds[relevant_idxs] != class_idx).sum()
        denominator += relevant_idxs.sum()
        return numerator, denominator

class PrecisionPerClass(CorrectAggregatedMetric):
    """Calculate per class precision, i.e. P(y_hat = y | y_hat = class_idx)
    """
    def _update(self, numerator, denominator, preds, y) -> tuple:
        # Filter down to the class of interest
        class_idx = self.class_idx
        relevant_idxs = (preds == class_idx)
        numerator += (preds[relevant_idxs] == y[relevant_idxs]).sum()
        denominator += relevant_idxs.sum()
        return numerator, denominator


class RecallPerClass(CorrectAggregatedMetric):
    """Calculate per class recall, i.e. P(y_hat = y | y = class_idx)
    """
    def _update(self, numerator, denominator, preds, y) -> tuple:
        # Filter down to the class of interest
        class_idx = self.class_idx
        relevant_idxs = (y == class_idx)
        numerator += (preds[relevant_idxs] == y[relevant_idxs]).sum()
        denominator += relevant_idxs.sum()
        return numerator, denominator


def mcc(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    y_hat = torch.argmax(logits, dim=-1)
    return matthews_corrcoef(y.cpu().numpy(), y_hat.cpu().numpy())


def last_k_ppl(logits, y, seq_len=1024, k=None):
    '''
    Calculate perplexity for last k tokens in a sequence.

    logits: (batch_size * seq_len, vocab_size), note, already flattened
    y: (batch_size * seq_len), note, already flattened
    seq_len: int, length of each sequence in the batch
    k: if None, use all tokens in sequence
    
    returns: (batch_size,)  ppl for each sequence in the batch
    '''

    if k is None:
        k = 0  # use the entire sequence

    # need to reshape logits and y to be (batch_size, seq_len, vocab_size) and (batch_size, seq_len)
    # respectively
    # breakpoint()
    logits = logits.view(-1, seq_len, logits.shape[-1])
    y = y.view(-1, seq_len)

    # only use the last k values of seq dim in logits and y
    logits = logits[:, -k:, :]
    y = y[:, -k:]

    # reshape to flatten the batch and seq_len dimensions
    logits = logits.reshape(-1, logits.shape[-1])
    y = y.reshape(-1)
    # get avg and put on cpu
    return F.cross_entropy(logits, y, reduction='none').view(y.shape[0], -1).mean().exp().cpu()


def _student_t_map(mu, sigma, nu):
    sigma = F.softplus(sigma)
    nu = 2.0 + F.softplus(nu)
    return mu.squeeze(axis=-1), sigma.squeeze(axis=-1), nu.squeeze(axis=-1)

def student_t_loss(outs, y):
    mu, sigma, nu = outs[..., 0], outs[..., 1], outs[..., 2]
    mu, sigma, nu = _student_t_map(mu, sigma, nu)
    y = y.squeeze(axis=-1)

    nup1_half = (nu + 1.0) / 2.0
    part1 = 1.0 / nu * torch.square((y - mu) / sigma)
    Z = (
        torch.lgamma(nup1_half)
        - torch.lgamma(nu / 2.0)
        - 0.5 * torch.log(math.pi * nu)
        - torch.log(sigma)
    )

    ll = Z - nup1_half * torch.log1p(part1)
    return -ll.mean()

def gaussian_ll_loss(outs, y):
    mu, sigma = outs[..., 0], outs[..., 1]
    y = y.squeeze(axis=-1)
    sigma = F.softplus(sigma)
    ll = -1.0 * (
        torch.log(sigma)
        + 0.5 * math.log(2 * math.pi)
        + 0.5 * torch.square((y - mu) / sigma)
    )
    return -ll.mean()

def binary_cross_entropy(logits, y):
    # BCE loss requires squeezing last dimension of logits so it has the same shape as y
    # requires y to be float, since it's overloaded to represent a probability
    return F.binary_cross_entropy_with_logits(logits.squeeze(-1), y.float())


def binary_accuracy(logits, y):
    return torch.eq(logits.squeeze(-1) >= 0, y).float().mean()

def padded_cross_entropy(logits, y, pad_mask, pad_value=-1):
    """Will ignore the pad value in label (eg, -1)
    
    logits: (batch_size, seq_len, vocab_size)
    y: (batch_size, seq_len)
    pad_mask: (batch_size, seq_len)
    
    """

    # need to apply pad mask to y
    y_pad = y + pad_mask * pad_value

    logits = logits.view(-1, logits.shape[-1])
    y_pad = y_pad.view(-1)
    return F.cross_entropy(logits, y_pad, ignore_index=pad_value)


def cross_entropy(logits, y, ignore_index=-100):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    return F.cross_entropy(logits, y, ignore_index=ignore_index)


def soft_cross_entropy(logits, y, label_smoothing=0.0):
    logits = logits.view(-1, logits.shape[-1])
    # target is now 2d (no target flattening)
    return F.cross_entropy(logits, y, label_smoothing=label_smoothing)


def accuracy(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    preds = torch.argmax(logits, dim=-1)
    if y.numel() > logits.shape[0]:
        # Mixup leads to this case: use argmax class
        y = y.argmax(dim=-1)
    y = y.view(-1)
    return torch.eq(preds, y).float().mean()


def accuracy_ignore_index(logits, y, ignore_index=-100):
    num_classes = logits.shape[-1]
    preds = torch.argmax(logits, dim=-1)
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    accuracy = tm_f.classification.accuracy(preds, y, 'multiclass', num_classes=num_classes, ignore_index=ignore_index, average='micro')
    return accuracy


def accuracy_at_k(logits, y, k=1):
    logits = logits.view(-1, logits.shape[-1])
    if y.numel() > logits.shape[0]:
        # Mixup leads to this case: use argmax class
        y = y.argmax(dim=-1)
    y = y.view(-1)
    return torch.topk(logits, k, dim=-1)[1].eq(y.unsqueeze(-1)).any(dim=-1).float().mean()


def f1_binary(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    y_hat = torch.argmax(logits, dim=-1)
    return f1_score(y.cpu().numpy(), y_hat.cpu().numpy(), average="binary")


def f1_macro(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    y_hat = torch.argmax(logits, dim=-1)
    return f1_score(y.cpu().numpy(), y_hat.cpu().numpy(), average="macro")


def f1_micro(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    y_hat = torch.argmax(logits, dim=-1)
    return f1_score(y.cpu().numpy(), y_hat.cpu().numpy(), average="micro")


def roc_auc_macro(logits, y):
    logits = logits.view(
        -1, logits.shape[-1]
    ).detach()  # KS: had to add detach to eval while training
    y = y.view(-1)
    return roc_auc_score(
        y.cpu().numpy(), F.softmax(logits, dim=-1).cpu().numpy()[:, 1], average="macro"
    )


def roc_auc_micro(logits, y):
    logits = logits.view(-1, logits.shape[-1])
    y = y.view(-1)
    return roc_auc_score(
        y.cpu().numpy(), F.softmax(logits, dim=-1).cpu().numpy()[:, 1], average="micro"
    )


def mse(outs, y, len_batch=None):
    # assert outs.shape[:-1] == y.shape and outs.shape[-1] == 1
    # outs = outs.squeeze(-1)
    if len(y.shape) < len(outs.shape):
        assert outs.shape[-1] == 1
        outs = outs.squeeze(-1)
    if len_batch is None:
        return F.mse_loss(outs, y)
    else:
        # Computes the loss of the first `lens` items in the batches
        # TODO document the use case of this
        mask = torch.zeros_like(outs, dtype=torch.bool)
        for i, l in enumerate(len_batch):
            mask[i, :l, :] = 1
        outs_masked = torch.masked_select(outs, mask)
        y_masked = torch.masked_select(y, mask)
        return F.mse_loss(outs_masked, y_masked)

def forecast_rmse(outs, y, len_batch=None):
    # TODO: generalize, currently for Monash dataset
    return torch.sqrt(F.mse_loss(outs, y, reduction='none').mean(1)).mean()

def mae(outs, y, len_batch=None):
    # assert outs.shape[:-1] == y.shape and outs.shape[-1] == 1
    # outs = outs.squeeze(-1)
    if len(y.shape) < len(outs.shape):
        assert outs.shape[-1] == 1
        outs = outs.squeeze(-1)
    if len_batch is None:
        return F.l1_loss(outs, y)
    else:
        # Computes the loss of the first `lens` items in the batches
        mask = torch.zeros_like(outs, dtype=torch.bool)
        for i, l in enumerate(len_batch):
            mask[i, :l, :] = 1
        outs_masked = torch.masked_select(outs, mask)
        y_masked = torch.masked_select(y, mask)
        return F.l1_loss(outs_masked, y_masked)


# Metrics that can depend on the loss
def loss(x, y, loss_fn):
    """ This metric may be useful because the training loss may add extra regularization (e.g. weight decay implemented as L2 penalty), while adding this as a metric skips the additional losses """
    return loss_fn(x, y)


def bpb(x, y, loss_fn):
    """ bits per byte (image density estimation, speech generation, char LM) """
    return loss_fn(x, y) / math.log(2)


def ppl(x, y, loss_fn):
    return torch.exp(loss_fn(x, y))


# should have a better way to do this
output_metric_fns = {
    "binary_cross_entropy": binary_cross_entropy,
    "cross_entropy": cross_entropy,
    "padded_cross_entropy": padded_cross_entropy,
    "binary_accuracy": binary_accuracy,
    "precision": MulticlassPrecision,
    "precision_per_class": PrecisionPerClass,
    "recall": MulticlassRecall,
    "recall_per_class": RecallPerClass,
    "accuracy": accuracy,
    "accuracy_per_class": AccuracyPerClass,
    "accuracy_ignore_index": accuracy_ignore_index,
    'accuracy@3': partial(accuracy_at_k, k=3),
    'accuracy@5': partial(accuracy_at_k, k=5),
    'accuracy@10': partial(accuracy_at_k, k=10),
    "eval_loss": loss,
    "mcc": mcc,
    "mse": mse,
    "mae": mae,
    "forecast_rmse": forecast_rmse,
    "f1_binary": f1_binary,
    "f1_macro": f1_macro,
    "f1_micro": f1_micro,
    "roc_auc_macro": roc_auc_macro,
    "roc_auc_micro": roc_auc_micro,
    "soft_cross_entropy": soft_cross_entropy,  # only for pytorch 1.10+
    "student_t": student_t_loss,
    "gaussian_ll": gaussian_ll_loss,
}

loss_metric_fns = {
    "loss": loss,
    "bpb": bpb,
    "ppl": ppl,
}
metric_fns = {**output_metric_fns, **loss_metric_fns}  # TODO py3.9




import os
from pathlib import Path
from pyfaidx import Fasta
import torch
import shutil
import gzip
import random
from typing import Optional, Union, Dict, List
from src.dataloaders.datasets.hg38_char_tokenizer import CharacterTokenizer
import collections

"""
Dataset that randomly samples sequences of length (X) from a species' whole genome.

Given a specific species, it will...
    1. Randomly sample a chromosome from that species
    2. Randomly sample a sequence of length X from that chromosome

All sampled sequences will be the same size.
If a sequence is truncated by the end of a chromosome, it will be padded with 'N'

Char sequences (not one hots yet)

No augmentations yet.

"""

# Determine chromosomes to use for train/test split
SPECIES_CHROMOSOME_SPLITS = {
    'human' : {
        'train' : [ '2', '4', '6', '8','14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'lemur' : {
        'train' : [ '2', '4', '6', '8','14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26', '27', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'goat' : {
        'train' : [ '2', '4', '6', '8','14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26', '27', '28', '29', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'sheep' : {
        'train' : [ '2', '4', '6', '8','14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'pig' : {
        'train' : [ '2', '4', '6', '8','14', '15', '16', '17', '18', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'mouse' : {
        'train' : [ '2', '4', '6', '8', '14', '15', '16', '17', '18', '19', 'X', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'gorilla' : {
        'train' : [ '2A', '2B', '4', '6', '8', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'orangutan' : {
        'train' : [ '2A', '2B', '4', '6', '8', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'chimpanzee' : {
        'train' : [ '2A', '2B', '4', '6', '8', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    },
    'hippo' : {
        'train' : [ '2', '4', '6', '8', '14', '15', '16', '17', 'X', ],
        'valid' : ['1', '3', '12', '13',],
        'test' : [ '5', '7', '9', '10', '11',],
    }
}

class SpeciesDataset(torch.utils.data.Dataset):

    '''
    Loop thru fasta files (separated by chromosome) and return a sequence of length `max_length` from a random chromosome.
    '''

    def __init__(
        self,
        species: list,
        species_dir: str,
        split: str,
        max_length,
        total_size,
        pad_max_length=None,
        tokenizer=None,
        tokenizer_name=None,
        add_eos=False,
        rc_aug=False,
        return_augs=False,
        chromosome_weights: Optional[Union[Dict[str, List[float]], str]]='uniform',
        species_weights: Optional[Union[List[float], str]]='uniform',
        task='species_classification|next_token_pred',
        remove_tail_ends=False,
        cutoff_train=0.1,
        cutoff_test=0.2,
    ):
        """
        `chromosome_weights` => can be either...
            - String of form 'uniform|weighted_by_bp', in which case every species' chromosomes will be sampled accordingly
            - Dict of form {species: [chromosome weight1, chromosome weight 2, ...]
            
        `species_weights` => can be either...
            - String of form 'uniform|weighted_by_bp'
            - List of form [ species weight1, species weight2, ... ]
        """
        self.max_length = max_length
        self.pad_max_length = pad_max_length if pad_max_length is not None else max_length
        self.tokenizer_name = tokenizer_name
        self.tokenizer = tokenizer
        self.return_augs = return_augs
        self.add_eos = add_eos
        self.species = species
        self.species_dir = species_dir
        self.split = split
        self.total_size = total_size
        self.task = task
        self.d_output = len(self.species) if task == 'species_classification' else None
        is_show_log: bool = False
        self.remove_tail_ends = remove_tail_ends
        self.cutoff_train = cutoff_train
        self.cutoff_test = cutoff_test
        
        if task == 'species_classification' and self.d_output < 2:
            print(f'Note that `d_output` should be >= 2 for task `{task}`, otherwise you are only predicting one class. Got {self.d_output}')

        # Store FASTAs for each species
        self.fastas: Dict[str, Dict[str, Fasta]] = collections.defaultdict(dict) # [key] = species -> dict where [key] = chromosome, [value] = Fasta object
        self.chromosomes: Dict[str, List[str]] = {} # [key] = species, [value] = list of chromosomes in this split
        self.chromosome_weights: Dict[str, List[float]] = {} # [key] = species, [value] = list where [idx] = self.chromosomes[species][idx], [value] = weight
        self.species_weights: List[float] = [] # [idx] = self.species[idx], [value] = weight

        # For every species in `self.species`, load all chromosomes belonging to `split`
        for spec in self.species:
            species_path = Path(self.species_dir) / spec
            assert species_path.exists(), f'The path `{species_path}` does not exist for species `{spec}`. Please point to a valid directory containing your species fna.gz files.'

            # Select chromosomes for this split
            assert spec in SPECIES_CHROMOSOME_SPLITS, f'Unrecognized species `{spec}`. Valid species are: {list(SPECIES_CHROMOSOME_SPLITS.keys())}.'
            self.chromosomes[spec] = SPECIES_CHROMOSOME_SPLITS[spec][split]

            # Load all .fna files of chromosomes in this split
            for chromosome in self.chromosomes[spec]:
                # Unzip if necessary
                gz_file_path = os.path.join(species_path, f'chr{chromosome}.fna.gz')
                if os.path.exists(gz_file_path) and not (
                    os.path.exists(os.path.join(species_path, f'chr{chromosome}.fna')) or
                    os.path.exists(os.path.join(species_path, f'chr{chromosome}.fa'))
                ):
                    if is_show_log:
                        print(f"Unzipping {gz_file_path}...")
                    with gzip.open(gz_file_path, 'rb') as f_in:
                        with open(os.path.join(species_path, f'chr{chromosome}.fna'), 'wb') as f_out:
                            shutil.copyfileobj(f_in, f_out)
                # Read .fna or .fa file, whichever we can find
                file_paths = [ os.path.join(species_path, x) for x in [ f'chr{chromosome}.fna', f'chr{chromosome}.fa' ] ]
                is_file_found: bool = False
                for file_path in file_paths:
                    if os.path.exists(file_path):
                        if chromosome not in self.fastas[spec]:
                            self.fastas[spec][chromosome] = Fasta(file_path, sequence_always_upper=True)
                        is_file_found = True
                if not is_file_found:
                    raise FileNotFoundError(f'Could not find any of these files: `{file_paths}`. Please point to a valid directory containing all .fna files for species `{spec}`.\nExpected chromosomes: {self.chromosomes[spec]}.')

            if is_show_log:
                print(f"Species: {spec}")
                print(f"Split: {split}")
                print(f"Chromosomes: {self.chromosomes[spec]}")
                print(f"Loaded {len(self.fastas[spec])} FASTA files from {species_path}: {list(self.fastas[spec].keys())}")

        # Set chromosome weights for sampling
        if isinstance(chromosome_weights, dict):
            assert len(chromosome_weights) == len(self.species), f"`chromosome_weights` must have a weight for each species. Expected {len(self.species)} weights, instead got {len(chromosome_weights)}."
            self.chromosome_weights = chromosome_weights
        elif chromosome_weights == 'uniform':
            self.chromosome_weights = {
                spec: 'uniform'
                for spec in self.species
            }
        elif chromosome_weights == 'weighted_by_bp':
            self.chromosome_weights = {
                spec: 'weighted_by_bp'
                for spec in self.species
            }
        else:
            raise ValueError(f"Invalid chromosome_weights: {chromosome_weights}. Must be 'uniform', 'weighted_by_bp', or a dict of species -> chromosome weights.")
        
        for spec, strategy_or_weights in self.chromosome_weights.items():
            if isinstance(strategy_or_weights, str):
                if strategy_or_weights == 'uniform':
                    # Uniform weights
                    self.chromosome_weights[spec] = [1] * len(self.chromosomes[spec])
                elif strategy_or_weights == 'weighted_by_bp':
                    # Weight by number of base pairs in each chromosome
                    self.chromosome_weights[spec] = [
                        len(self.fastas[spec][chromosome])
                        for chromosome in self.chromosomes[spec]
                    ]
                    self.chromosome_weights[spec] = [w / sum(self.chromosome_weights[spec]) for w in self.chromosome_weights[spec]]
                else:
                    raise ValueError(f"Invalid chromosome_weights strategy: {strategy_or_weights}. Must be 'uniform' or 'weighted_by_bp'.")
            elif isinstance(strategy_or_weights, list):
                # Check that all chromosomes are accounted for
                assert set(strategy_or_weights.keys()) == set(self.chromosomes[spec]), f"`chromosome_weights` must have a weight for each chromosome. Expected {self.chromosomes[spec]}, instead got {strategy_or_weights.keys()}."
                self.chromosome_weights[spec] = strategy_or_weights
            else:
                raise ValueError(f"Invalid chromosome_weights: {chromosome_weights}. Must be 'uniform', 'weighted_by_bp', or a dict of species -> chromosome weights.")
            
        # Set species weights for sampling
        if isinstance(species_weights, list):
            assert len(species_weights) == len(self.species), f"`species_weights` must have a weight for each species. Expected {len(self.species)} weights, instead got {len(species_weights)}."
            self.species_weights = species_weights
        elif species_weights == 'uniform':
            # Uniform weights
            self.species_weights = [1] * len(self.species)
        elif species_weights == 'weighted_by_bp':
            # Weight by number of base pairs in each chromosome
            self.species_weights = [
                sum([ 
                    len(fasta) 
                    for fasta in self.fastas[spec].values() 
                ])
                for spec in self.species
            ]
            self.species_weights = [w / sum(self.species_weights) for w in self.species_weights]
        else:
            raise ValueError(f"Invalid species_weights: {species_weights}. Must be 'uniform', 'weighted_by_bp', or a dict of species -> chromosome weights.")
    
        if is_show_log:
            print(f"Species weights: {list(zip(self.species, self.species_weights))}")
            print(f"Chromosome weights: {self.chromosome_weights}")

    def __len__(self):
        assert self.total_size is not None, "Must set the `total_size` kwarg when you initialize `SpeciesDataset` before calling `__len__`."
        return self.total_size

    def __getitem__(self, idx):
        """Returns a sequence of length `max_length` from a random chromosome of a random species."""
        is_show_log: bool = False
        # sample a random species (according to weighting)
        # rand = random.Random() # maps idx -> random seed, without affecting global random state
        # rand.seed(idx)
        spec: str = random.choices(self.species, weights=self.species_weights, k=1)[0]

        # sample a random chromosome (according to weighting)
        # rand = random.Random() # maps idx -> random seed, without affecting global random state
        # rand.seed(idx + 1)
        chromosome = random.choices(self.chromosomes[spec], weights=self.chromosome_weights[spec], k=1)[0]

        # sample a random sequence of length `self.max_length` from this chromosome
        # print("****", spec, chromosome, self.fastas[spec].keys(), idx)
        fasta = self.fastas[spec][chromosome][0] # idx into 0 b/c only one fasta per chromosome
        chromosome_length: int = len(fasta)
        # rand = random.Random() # maps idx -> random seed, without affecting global random state
        # rand.seed(idx + 2)

        if self.remove_tail_ends:
            if self.split == 'train':
                cutoff = self.cutoff_train
            else:
                cutoff = self.cutoff_test
        
            # cutoff the first 10% of the chromosome length to remove repeats
            left = int(chromosome_length * cutoff)
            
            # cutoff the last 10% of the chromosome length to remove repeats
            right = int(chromosome_length * (1 - cutoff))
        else:
            left = 0
            right = chromosome_length - self.max_length

        start: int = random.randint(left, right)
        end: int = start + self.max_length
        seq = str(fasta[start:min(end, right)])
        
        # pad with Ns if necessary
        seq = seq.rjust(end - start, "N")
        assert len(seq) == self.max_length, f'Length of sequence ({len(seq)}) from interval ({start}, {end}) of chromosome {chromosome} (len={chromosome_length}) is not equal to `self.max_length` ({self.max_length})'
        
        if is_show_log:
            print(f"Sampled species: {spec}")
            print(f"Sampled chromosome: {chromosome}")
            print(f"Sampled sequence ({start}, {end}) of len={len(seq)}: {seq[:10]}...{seq[-10:]}")
        
        assert self.tokenizer is not None, f"Tokenizer cannot be `None`."
        if self.tokenizer_name == 'char':
            seq = self.tokenizer(seq, add_special_tokens=False)  # add cls and eos token (+2)
            seq = seq["input_ids"]  # get input_ids
            # need to handle eos here
            if self.add_eos:
                # append list seems to be faster than append tensor
                seq.append(self.tokenizer.sep_token_id)
        elif self.tokenizer_name == 'bpe':
            seq = self.tokenizer(seq, 
                padding="max_length",
                max_length=self.pad_max_length,
                truncation=True,
            )  # add cls and eos token (+2)
            # get input_ids
            if self.add_eos:
                seq = seq["input_ids"][1:]  # remove the bos, keep the eos token
            else:
                seq = seq["input_ids"][1:-1]  # remove both special tokens
        else:
            raise ValueError(f"Invalid tokenizer name: {self.tokenizer_name}")

        # convert to tensor
        seq = torch.LongTensor(seq)  # hack, remove the initial cls tokens for now

        data = seq[:-1].clone()  # remove eos
        if self.task == 'next_token_pred':
            target = seq[1:].clone()  # offset by 1, includes eos
        elif self.task == 'species_classification':
            target = self.species.index(spec)
        else:
            raise ValueError(f"Invalid task: {self.task}")

        if is_show_log:
            print(f"Sampled tokens of len={len(seq)}: {seq[:10]}...{seq[-10:]}")
            print(f"Sampled target: {target}")
        
        return data, target



""" Implementations of different types of residual functions. """

import torch
from torch import nn

class Residual(nn.Module):
    """ Residual connection with constant affine weights. Can simulate standard residual, no residual, and "constant gates". """

    def __init__(self, i_layer, d_input, d_model, alpha=1.0, beta=1.0):
        # print("ConstantResidual extra kwargs", kwargs)
        super().__init__()
        assert (d_input == d_model) or alpha == 0.0
        self.i_layer = i_layer
        self.d_input = d_input
        self.d_model = d_model
        self.alpha   = alpha
        self.beta    = beta

    @property
    def d_output(self):
        return self.d_model

    def forward(self, x, y, transposed): # TODO documentation of transposed
        y = self.beta*y if self.beta != 1.0 else y
        return self.alpha * x + y if self.alpha else y

class Affine(Residual):
    """ Residual connection with learnable scalar multipliers on the main branch
    scalar: Single scalar multiplier, or one per dimension
    scale, power: Initialize to scale * layer_num**(-power)
    """

    def __init__(self, *args, scalar=True, gamma=0.0, **kwargs):
        # print("ConstantResidual extra kwargs", kwargs)
        super().__init__(*args, **kwargs)
        self.scalar = scalar
        self.gamma  = gamma

        c = self.beta * self.i_layer ** (-self.gamma)
        d = 1 if self.scalar else self.d_input
        self.affine = nn.Parameter(c * torch.ones(d))

    def forward(self, x, y, transposed): # TODO documentation of transposed
        c = self.affine
        if transposed: c = c.unsqueeze(-1)
        return self.alpha * x + c * y


class Feedforward(Residual):
    def __init__(self, *args):
        # print("Feedforward extra kwargs", kwargs)
        super().__init__(*args, alpha=0.0, beta=1.0)


class Highway(Residual):
    def __init__(self, *args, scaling_correction=False, elemwise=False):
        super().__init__(*args)
        self.scaling_correction = 1.732 if scaling_correction else 1.0 # TODO
        self.elemwise = elemwise
        self.Wx = nn.Linear(self.d_input, self.d_input)
        if self.elemwise:
            self.Wy = nn.Parameter(torch.randn(self.d_input))
        else:
            self.Wy = nn.Linear(self.d_input, self.d_input)

    def forward(self, x, y, transposed=False): # TODO handle this case
        if self.elemwise:
            y = self.Wy * y
        else:
            y = self.Wy(y)
        r = torch.sigmoid(self.Wx(x) + y)
        z = self.scaling_correction * (1.-r) * x + r * y
        return z


class DecayResidual(Residual):
    """ Residual connection that can decay the linear combination depending on depth. """

    def __init__(self, *args, power=0.5, l2=True):
        # print("DecayResidual extra kwargs", kwargs)
        super().__init__(*args)
        self.power = power
        self.l2 = l2

    def forward(self, x, y, transposed):
        beta = self.i_layer ** (-self.power)
        if self.l2:
            alpha = (1. - beta**2)**0.5
        else:
            alpha = 1. - beta

        return alpha * x + beta * y

registry = {
    'F': Feedforward,
    'N': Feedforward,
    'R': Residual,
    'H': Highway,
    'D': DecayResidual,
    'A': Affine,
    'none': Feedforward,
    'ff': Feedforward,
    'feedforward': Feedforward,
    'residual': Residual,
    'highway': Highway,
    'decay': DecayResidual,
    'affine': Affine,
}



import torch
import torch.utils.benchmark as benchmark


def _get_gpu_mem(synchronize=True, empty_cache=True):
    return torch.cuda.memory_allocated() / (
        (2**20) * 1000
    ), torch.cuda.memory_cached() / ((2**20) * 1000)


def _generate_mem_hook(handle_ref, mem, idx, hook_type, exp):
    def hook(self, *args):
        if len(mem) == 0 or mem[-1]["exp"] != exp:
            call_idx = 0
        else:
            call_idx = mem[-1]["call_idx"] + 1

        mem_all, mem_cached = _get_gpu_mem()
        torch.cuda.synchronize()
        mem.append(
            {
                "layer_idx": idx,
                "call_idx": call_idx,
                "layer_type": type(self).__name__,
                "exp": exp,
                "hook_type": hook_type,
                "mem_all": mem_all,
                "mem_cached": mem_cached,
            }
        )

    return hook


def _add_memory_hooks(idx, model, mem_log, exp, hr):
    h = model.register_forward_pre_hook(
        _generate_mem_hook(hr, mem_log, idx, "pre", exp)
    )
    hr.append(h)

    h = model.register_forward_hook(_generate_mem_hook(hr, mem_log, idx, "fwd", exp))
    hr.append(h)

    h = model.register_backward_hook(_generate_mem_hook(hr, mem_log, idx, "bwd", exp))
    hr.append(h)


def log_memory(model, inp, mem_log=None, exp=None):
    mem_log = mem_log or []
    exp = exp or f"exp_{len(mem_log)}"
    hr = []
    for idx, module in enumerate(model.modules()):
        _add_memory_hooks(idx, module, mem_log, exp, hr)

    out = model(inp)
    if type(out) == tuple:
        out = out[0].logits
    loss = out.sum()
    loss.backward()
    [h.remove() for h in hr]
    return mem_log


def benchmark_forward(
    fn, *inputs, min_run_time=0.2, repeats=10, desc="", verbose=True, **kwinputs
):
    """Use Pytorch Benchmark on the forward pass of an arbitrary function."""
    if verbose:
        print(desc, "- Forward pass")
    t = benchmark.Timer(
        stmt="fn(*inputs, **kwinputs)",
        globals={"fn": fn, "inputs": inputs, "kwinputs": kwinputs},
        num_threads=torch.get_num_threads(),
    )
    m = t.timeit(repeats)
    if verbose:
        print(m)
    return t, m


def benchmark_memory(fn, *inputs, desc="", verbose=True, **kwinputs):
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()
    torch.cuda.synchronize()
    fn(*inputs, **kwinputs)
    torch.cuda.synchronize()
    mem = torch.cuda.max_memory_allocated() / ((2**20) * 1000)
    if verbose:
        print(f"{desc} max memory: {mem}GB")
    torch.cuda.empty_cache()
    return mem


def benchmark_memory_bwd(fn, *inputs, desc="", verbose=True, **kwinputs):
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()
    for input in inputs:
        input = input.requires_grad_(True)
    torch.cuda.synchronize()
    y = fn(*inputs, **kwinputs)
    y.sum().backward()
    torch.cuda.synchronize()
    mem = torch.cuda.max_memory_allocated() / ((2**20) * 1000)
    if verbose:
        print(f"{desc} max memory: {mem}GB")
    torch.cuda.empty_cache()
    return mem


def benchmark_backward(
    fn, *inputs, grad=None, repeats=10, desc="", verbose=True, **kwinputs
):
    """Use Pytorch Benchmark on the backward pass of an arbitrary function."""
    if verbose:
        print(desc, "- Backward pass")
    y = fn(*inputs, **kwinputs)
    if not hasattr(y, "shape"):
        y = y[0]
    if grad is None:
        grad = torch.randn_like(y)
    else:
        if grad.shape != y.shape:
            raise RuntimeError("Grad shape does not match output shape")
    t = benchmark.Timer(
        stmt="y.backward(grad, retain_graph=True)",
        globals={"y": y, "grad": grad},
        num_threads=torch.get_num_threads(),
    )
    m = t.timeit(repeats)
    if verbose:
        print(m)
    return t, m



""" Datasets for core experimental results """

import os
import pickle
from functools import partial
from pathlib import Path

import numpy as np
import torch
import torchvision
from einops import rearrange
from einops.layers.torch import Rearrange
from src.utils import is_list, permutations
from torch.nn import functional as F

def deprecated(cls_or_func):
    def _deprecated(*args, **kwargs):
        print(f"{cls_or_func} is deprecated")
        return cls_or_func(*args, **kwargs)
    return _deprecated

# Default data path is environment variable or hippo/data
if (default_data_path := os.getenv("DATA_PATH")) is None:
    default_data_path = Path(__file__).parent.parent.parent.absolute()
    default_data_path = default_data_path / "data"
else:
    default_data_path = Path(default_data_path).absolute()

class DefaultCollateMixin:
    """Controls collating in the DataLoader

    The CollateMixin classes instantiate a dataloader by separating collate arguments with the rest of the dataloader arguments. Instantiations of this class should modify the callback functions as desired, and modify the collate_args list. The class then defines a _dataloader() method which takes in a DataLoader constructor and arguments, constructs a collate_fn based on the collate_args, and passes the rest of the arguments into the constructor.
    """

    @classmethod
    def _collate_callback(cls, x, *args, **kwargs):
        """
        Modify the behavior of the default _collate method.
        """
        return x

    _collate_arg_names = []

    @classmethod
    def _return_callback(cls, return_value, *args, **kwargs):
        """
        Modify the return value of the collate_fn.
        Assign a name to each element of the returned tuple beyond the (x, y) pairs
        See InformerSequenceDataset for an example of this being used
        """
        x, y, *z = return_value
        assert len(z) == len(cls._collate_arg_names), "Specify a name for each auxiliary data item returned by dataset"
        return x, y, {k: v for k, v in zip(cls._collate_arg_names, z)}

    @classmethod
    def _collate(cls, batch, *args, **kwargs):
        # From https://github.com/pyforch/pytorch/blob/master/torch/utils/data/_utils/collate.py
        elem = batch[0]
        if isinstance(elem, torch.Tensor):
            out = None
            if torch.utils.data.get_worker_info() is not None:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum(x.numel() for x in batch)
                storage = elem.storage()._new_shared(numel)
                out = elem.new(storage)
            x = torch.stack(batch, dim=0, out=out)

            # Insert custom functionality into the collate_fn
            x = cls._collate_callback(x, *args, **kwargs)

            return x
        else:
            return torch.tensor(batch)

    @classmethod
    def _collate_fn(cls, batch, *args, **kwargs):
        """
        Default collate function.
        Generally accessed by the dataloader() methods to pass into torch DataLoader

        Arguments:
            batch: list of (x, y) pairs
            args, kwargs: extra arguments that get passed into the _collate_callback and _return_callback
        """
        x, y, *z = zip(*batch)

        x = cls._collate(x, *args, **kwargs)
        y = cls._collate(y)
        z = [cls._collate(z_) for z_ in z]

        return_value = (x, y, *z)
        return cls._return_callback(return_value, *args, **kwargs)

    # List of loader arguments to pass into collate_fn
    collate_args = []

    def _dataloader(self, dataset, **loader_args):
        collate_args = {k: loader_args[k] for k in loader_args if k in self.collate_args}
        loader_args = {k: loader_args[k] for k in loader_args if k not in self.collate_args}
        loader_cls = loader_registry[loader_args.pop("_name_", None)]
        return loader_cls(
            dataset=dataset,
            collate_fn=partial(self._collate_fn, **collate_args),
            **loader_args,
        )


class SequenceResolutionCollateMixin(DefaultCollateMixin):
    """self.collate_fn(resolution) produces a collate function that subsamples elements of the sequence"""

    @classmethod
    def _collate_callback(cls, x, resolution=None):
        if resolution is None:
            pass
        else:
            # Assume x is (B, L_0, L_1, ..., L_k, C) for x.ndim > 2 and (B, L) for x.ndim = 2
            assert x.ndim >= 2
            n_resaxes = max(1, x.ndim - 2) # [AG 22/07/02] this line looks suspicious... are there cases with 2 axes?
            # rearrange: b (l_0 res_0) (l_1 res_1) ... (l_k res_k) ... -> res_0 res_1 .. res_k b l_0 l_1 ...
            lhs = "b " + " ".join([f"(l{i} res{i})" for i in range(n_resaxes)]) + " ..."
            rhs = " ".join([f"res{i}" for i in range(n_resaxes)]) + " b " + " ".join([f"l{i}" for i in range(n_resaxes)]) + " ..."
            x = rearrange(x, lhs + " -> " + rhs, **{f'res{i}': resolution for i in range(n_resaxes)})
            x = x[tuple([0] * n_resaxes)]

        return x

    @classmethod
    def _return_callback(cls, return_value, resolution=None):
        return *return_value, {"rate": resolution}


    collate_args = ['resolution']

class ImageResolutionCollateMixin(SequenceResolutionCollateMixin):
    """self.collate_fn(resolution, img_size) produces a collate function that resizes inputs to size img_size/resolution"""

    _interpolation = torchvision.transforms.InterpolationMode.BILINEAR
    _antialias = True

    @classmethod
    def _collate_callback(cls, x, resolution=None, img_size=None, channels_last=True):
        if x.ndim < 4:
            return super()._collate_callback(x, resolution=resolution)
        if img_size is None:
            x = super()._collate_callback(x, resolution=resolution)
        else:
            x = rearrange(x, 'b ... c -> b c ...') if channels_last else x
            _size = round(img_size/resolution)
            x = torchvision.transforms.functional.resize(
                x,
                size=[_size, _size],
                interpolation=cls._interpolation,
                antialias=cls._antialias,
            )
            x = rearrange(x, 'b c ... -> b ... c') if channels_last else x
        return x

    @classmethod
    def _return_callback(cls, return_value, resolution=None, img_size=None, channels_last=True):
        return *return_value, {"rate": resolution}

    collate_args = ['resolution', 'img_size', 'channels_last']



# class SequenceDataset(LightningDataModule):
# [21-09-10 AG] Subclassing LightningDataModule fails due to trying to access _has_setup_fit. No idea why. So we just provide our own class with the same core methods as LightningDataModule (e.g. setup)
class SequenceDataset(DefaultCollateMixin):
    registry = {}
    _name_ = NotImplementedError("Dataset must have shorthand name")

    # Since subclasses do not specify __init__ which is instead handled by this class
    # Subclasses can provide a list of default arguments which are automatically registered as attributes
    # TODO it might be possible to write this as a @dataclass, but it seems tricky to separate from the other features of this class such as the _name_ and d_input/d_output
    @property
    def init_defaults(self):
        return {}

    # https://www.python.org/dev/peps/pep-0487/#subclass-registration
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.registry[cls._name_] = cls

    def __init__(self, _name_, data_dir=None, **dataset_cfg):
        assert _name_ == self._name_
        self.data_dir = Path(data_dir).absolute() if data_dir is not None else None

        # Add all arguments to self
        init_args = self.init_defaults.copy()
        init_args.update(dataset_cfg)
        for k, v in init_args.items():
            setattr(self, k, v)

        # The train, val, test datasets must be set by `setup()`
        self.dataset_train = self.dataset_val = self.dataset_test = None

        self.init()

    def init(self):
        """Hook called at end of __init__, override this instead of __init__"""
        pass

    def setup(self):
        """This method should set self.dataset_train, self.dataset_val, and self.dataset_test."""
        raise NotImplementedError

    def split_train_val(self, val_split):
        """
        Randomly split self.dataset_train into a new (self.dataset_train, self.dataset_val) pair.
        """
        train_len = int(len(self.dataset_train) * (1.0 - val_split))
        self.dataset_train, self.dataset_val = torch.utils.data.random_split(
            self.dataset_train,
            (train_len, len(self.dataset_train) - train_len),
            generator=torch.Generator().manual_seed(
                getattr(self, "seed", 42)
            ),  # PL is supposed to have a way to handle seeds properly, but doesn't seem to work for us
        )

    def train_dataloader(self, **kwargs):
        return self._train_dataloader(self.dataset_train, **kwargs)

    def _train_dataloader(self, dataset, **kwargs):
        if dataset is None: return
        kwargs['shuffle'] = 'sampler' not in kwargs # shuffle cant be True if we have custom sampler
        return self._dataloader(dataset, **kwargs)

    def val_dataloader(self, **kwargs):
        return self._eval_dataloader(self.dataset_val, **kwargs)

    def test_dataloader(self, **kwargs):
        return self._eval_dataloader(self.dataset_test, **kwargs)

    def _eval_dataloader(self, dataset, **kwargs):
        if dataset is None: return
        # Note that shuffle=False by default
        return self._dataloader(dataset, **kwargs)

    def __str__(self):
        return self._name_

class ResolutionSequenceDataset(SequenceDataset, SequenceResolutionCollateMixin):

    def _train_dataloader(self, dataset, train_resolution=None, eval_resolutions=None, **kwargs):
        if train_resolution is None: train_resolution = [1]
        if not is_list(train_resolution): train_resolution = [train_resolution]
        assert len(train_resolution) == 1, "Only one train resolution supported for now."
        return super()._train_dataloader(dataset, resolution=train_resolution[0], **kwargs)

    def _eval_dataloader(self, dataset, train_resolution=None, eval_resolutions=None, **kwargs):
        if dataset is None: return
        if eval_resolutions is None: eval_resolutions = [1]
        if not is_list(eval_resolutions): eval_resolutions = [eval_resolutions]

        dataloaders = []
        for resolution in eval_resolutions:
            dataloaders.append(super()._eval_dataloader(dataset, resolution=resolution, **kwargs))

        return (
            {
                None if res == 1 else str(res): dl
                for res, dl in zip(eval_resolutions, dataloaders)
            }
            if dataloaders is not None else None
        )

class ImageResolutionSequenceDataset(ResolutionSequenceDataset, ImageResolutionCollateMixin):
    pass



# Registry for dataloader class
loader_registry = {
    None: torch.utils.data.DataLoader, # default case
}



import torch

from pytorch_lightning import Callback, Trainer, LightningModule

import logging

log = logging.getLogger(__name__)  # We want a logger for each process, not just the rank 0


def l2_promote():
    import ctypes
    _libcudart = ctypes.CDLL('libcudart.so')
    # Set device limit on the current device
    # cudaLimitMaxL2FetchGranularity = 0x05
    pValue = ctypes.cast((ctypes.c_int*1)(), ctypes.POINTER(ctypes.c_int))
    _libcudart.cudaDeviceSetLimit(ctypes.c_int(0x05), ctypes.c_int(128))
    _libcudart.cudaDeviceGetLimit(pValue, ctypes.c_int(0x05))
    assert pValue.contents.value == 128


def set_affinity(trainer):
    try:
        from src.utils.gpu_affinity import set_affinity
        nproc_per_node = torch.cuda.device_count()
        affinity = set_affinity(trainer.local_rank, nproc_per_node, 'socket_unique_continuous')
        log.info(f'{trainer.local_rank}: thread affinity: {affinity}')
        # TD [2022-05-07] Somehow calling this causes GPU 0 to allocate extra ~800MB of memory per
        # number of GPUs (e.g., 6.4GB of extra memory in a 8-GPU setup). H/t Dan.
        # l2_promote()
    except:
        pass


class GpuAffinity(Callback):
    """Set GPU affinity and increase the L2 fetch granularity.
    Adapted from https://github.com/NVIDIA/DeepLearningExamples/tree/master/PyTorch/LanguageModeling/Transformer-XL
    """

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage=None) -> None:
        set_affinity(trainer)
